# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A moving average implementation."""
# from redipy.api import RedisClientAPI
# from redipy.backend.backend import ExecFunction
# from redipy.symbolic.seq import FnContext


# class RMovingAverage:
#     def __init__(self, rt: RedisClientAPI) -> None:
#         self._rt = rt
#         self._update = self._update_script()

#     def _update_script(self) -> ExecFunction:
#         ctx = FnContext()


#         rsize = RedisVar(ctx.add_key("size"))
#         base = ctx.add_local(ctx.add_key("frame"))
#         field = ctx.add_arg("field")
#         pos = ctx.add_local(ToNum(rsize.get_value(default=0)))
#         res = ctx.add_local(None)
#         cur = ctx.add_local(None)
#         rframe = RedisHash(cur)

#         loop = ctx.while_(res.eq_(None).and_(pos.ge_(0)))
#         loop.add(cur.assign(Strs(base, ":", ToIntStr(pos))))
#         loop.add(res.assign(rframe.hget(field)))
#         loop.add(pos.assign(pos - 1))

#         ctx.set_return_value(res)
#         return self._rt.register_script(ctx)



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Access to redis sets."""
from redipy.symbolic.expr import Expr, MixedType
from redipy.symbolic.fun import RedisObj


class RedisSet(RedisObj):
    """A redis set."""
    def add(self, value: MixedType) -> Expr:
        """
        Adds an element to the set.

        Args:
            value (MixedType): The value.

        Returns:
            Expr: The expression.
        """
        # FIXME add all arguments
        return self.redis_fn("sadd", value)

    def remove(self, value: MixedType) -> Expr:
        """
        Removes an element from the set.

        Args:
            value (MixedType): The value.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("srem", value)

    def has(self, value: MixedType) -> Expr:
        """
        Whether the set contains the given value.

        Args:
            value (MixedType): The value.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("sismember", value)

    def members(self) -> Expr:
        """
        Returns a list of set members.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("smembers")

    def card(self) -> Expr:
        """
        Computes the cardinality of the set.

        Returns:
            Expr: The expression.
        """
        return self.redis_fn("scard")



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module handles the internal state of the memory runtime."""
import bisect
import collections
import itertools
import threading
import time
from collections.abc import Callable, Iterable
from datetime import datetime
from typing import Literal, overload, TypeVar

from redipy.api import (
    KeyType,
    RedisAPI,
    REX_ALWAYS,
    REX_EARLIER,
    REX_EXPIRE,
    REX_LATER,
    REX_PERSIST,
    RExpireMode,
    RSetMode,
    RSM_ALWAYS,
    RSM_EXISTS,
    RSM_MISSING,
)
from redipy.util import (
    convert_pattern,
    now,
    reject_patterns,
    time_diff,
    to_number_str,
)


T = TypeVar('T')


def compute_expire(
        now_mono: float,
        now_ts: datetime,
        *,
        expire_timestamp: datetime | None,
        expire_in: float | None) -> float | None:
    """
    Computes the monotonic time point when a key should expire.

    Args:
        now_mono (float): The current monotonic time.
        now_ts (datetime): The current date time.
        expire_timestamp (datetime | None): An absolute timestamp.
        expire_in (float | None): A relative time difference in seconds.

    Raises:
        ValueError: If both expire_timestamp and expire_in are set.

    Returns:
        float | None: The monotonic time point.
    """
    if expire_timestamp is None:
        if expire_in is None:
            return None
        return now_mono + expire_in
    if expire_in is not None:
        raise ValueError(
            f"cannot set timestamp {expire_timestamp} "
            f"and duration {expire_in} at the same time")
    return now_mono + time_diff(now_ts, expire_timestamp)


MIN_SCAN_LENGTH: int = 10
"""The internal minimum number of items returned by `scan`."""


KEYS_MAX_SCAN: int = 10000
"""The internal maximum length for `scan` used by `keys`."""


class State:
    """
    The state holding the actual values of the memory runtime. A state can have
    a parent state which values it shadows. The parent can be updated by
    applying all changes.
    """
    def __init__(self, parent: 'State | None' = None) -> None:
        """
        Creates a new state.

        Args:
            parent (State | None, optional): The optional parent whose values
                this state shadows. Defaults to None.
        """
        super().__init__()
        self._parent = parent
        self._expire: dict[str, float] = {}
        self._vals: dict[str, str] = {}
        self._queues: dict[str, collections.deque[str]] = {}
        self._hashes: dict[str, dict[str, str]] = {}
        self._sets: dict[str, set[str]] = {}
        # TODO: to be replaced by something more efficient later
        self._zorder: dict[str, list[str]] = {}
        self._zscores: dict[str, dict[str, float]] = {}
        self._deletes: set[str] = set()
        self._key_cache: list[str] | None = None
        self._delete_count: int = 0

    def has_parent(self) -> bool:
        """
        Whether the state has a parent.

        Returns:
            bool: True, if this state has a parent.
        """
        return self._parent is not None

    def key_type(self, key: str) -> KeyType | None:
        """
        Computes the type of a given key. The function does not check
        expiration.

        Args:
            key (str): The key.

        Returns:
            KeyType | None: The type of the key or None if it doesn't exist.
        """
        if self._is_pending_delete(key):
            return None
        if key in self._vals:
            return "string"
        if key in self._queues:
            return "list"
        if key in self._hashes:
            return "hash"
        if key in self._sets:
            return "set"
        if key in self._zorder:
            return "zset"
        if self._parent is not None:
            return self._parent.key_type(key)
        return None

    def verify_key(self, key_type: KeyType, key: str) -> None:
        """
        Ensures that the given key does not already exist with a different
        type. This does not check for expiration.

        Args:
            key_type (KeyType): The expected key type.
            key (str): The key.

        Raises:
            TypeError: If the key already exists with a different type.
        """
        if self._is_pending_delete(key):
            return
        if key_type != "string" and key in self._vals:
            raise TypeError(f"key {key} is a string")
        if key_type != "list" and key in self._queues:
            raise TypeError(f"key {key} is a list")
        if key_type != "hash" and key in self._hashes:
            raise TypeError(f"key {key} is a hash")
        if key_type != "set" and key in self._sets:
            raise TypeError(f"key {key} is a set")
        if key_type != "zset" and key in self._zorder:
            raise TypeError(f"key {key} is a zset")
        if self._parent is not None:
            self._parent.verify_key(key_type, key)

    def _flush_key_cache(self) -> None:
        self._key_cache = None

    def _populate_key_cache(self, now_mono: float) -> list[str]:
        """
        Populates the key cache.

        Args:
            now_mono (float): The current time.

        Returns:
            list[str]: A sorted list of all keys.
        """
        self.clean_vals(now_mono)
        pre = [
            *self._vals,
            *self._queues,
            *self._hashes,
            *self._sets,
            *self._zorder,
        ]
        if self._parent is None:
            key_cache = sorted(pre)
        else:
            # NOTE: no need to remove duplicates here as they are allowed
            # by the spec and are removed downstream which is much cheaper
            parent_cache = self._parent.get_key_cache(now_mono)
            key_cache = sorted(pre + parent_cache)
        self._key_cache = key_cache
        return key_cache

    def get_key_cache(self, now_mono: float) -> list[str]:
        """
        Retrieves the full key cache. If necessary populates the key cache.
        The key cache is a sorted list of all keys. It is used for scanning
        keys.

        Args:
            now_mono (float): The current time.

        Returns:
            list[str]: A sorted list of all keys.
        """
        res = self._key_cache
        if res is None:
            res = self._populate_key_cache(now_mono)
        return res

    def _start_ix(self, pattern_prefix: str, now_mono: float) -> int:
        """
        Computes the start index in the key cache for a given pattern prefix.

        Args:
            pattern_prefix (str): The pattern prefix is the longest prefix of
                the pattern that does not include any wildcards or ranges.
            now_mono (float): The current time.

        Returns:
            int: Where the key cache can be started to be scanned to
                immediately retrieve keys starting with at least the pattern
                prefix.
        """
        return bisect.bisect_left(self.get_key_cache(now_mono), pattern_prefix)

    def _scan(
            self,
            cursor: int,
            length: int,
            pattern_prefix: str | None,
            now_mono: float) -> tuple[int, list[str]]:
        """
        Internal scan. The result might include keys that will be filtered in
        the actual scan function.

        Args:
            cursor (int): The scan cursor. If the scan cursor is 0 the starting
                position in the key cache is used. If the scan cursor is not 0
                it is the adjusted index one after the last returned key. The
                index is adjusted by the total amount of keys deleted. During
                iteration, if a key is added to the left of the cursor index
                it will not be returned (which is allowed by the spec) and some
                previously returned keys might be returned again. If the key is
                added to the right it will eventually be returned as well. If
                a key is deleted the adjustment of the cursor is shifted left
                by one (i.e., `delete_count` is increased by one). Due to the
                shift we continue iterating as if nothing happened. Shifting
                the index to the left guarantees that we do not skip an
                existing key if the deleted key was to the left.
            length (int): The length of the returned result. To avoid expensive
                iterations with small return lengths this value is clipped to
                `MIN_SCAN_LENGTH` on the lower end.
            pattern_prefix (str | None): The pattern prefix is the longest
                prefix of the pattern that does not include any wildcards or
                ranges. If None, all keys are considered.
            now_mono (float): The current time.

        Returns:
            tuple[int, list[str]]: A tuple of the next cursor and the returned
                keys. If the iteration is complete the next cursor is 0.
        """
        delete_count = self._delete_count
        if cursor == 0 and pattern_prefix is not None:
            cursor = self._start_ix(pattern_prefix, now_mono)
        else:
            cursor = max(cursor - delete_count, 0)
        key_cache = self.get_key_cache(now_mono)
        end_ix = max(cursor + max(length, MIN_SCAN_LENGTH), 1)
        res = key_cache[cursor:end_ix]
        if end_ix >= len(key_cache):
            end_ix = 0
        else:
            end_ix += delete_count
        return end_ix, res

    def scan_keys(
            self,
            cursor: int,
            now_mono: float,
            *,
            count: int,
            match: str | None,
            filter_type: KeyType | None) -> tuple[int, list[str]]:
        """
        Scan through all keys.

        Args:
            cursor (int): The scan cursor. A value of 0 initiates the beginning
                of the scan. All other values are defined internally and should
                only be previous return values of scan.
            now_mono (float): The current time.
            count (int): The length hint of the returned values. The actual
                number of returned keys might differ.
            match (str | None): If not None a redis key pattern to filter
                keys.
            filter_type (KeyType | None): If not None filter by key type.

        Returns:
            tuple[int, list[str]]: A tuple of the next cursor and the returned
                keys. If the iteration is complete the next cursor is 0.
        """
        if match is not None:
            prefix, pat = convert_pattern(match)
        else:
            prefix = None
            pat = None

        def process_type(keys: Iterable[str]) -> Iterable[str]:
            if filter_type is None:
                return keys
            return (
                key
                for key in keys
                if self.key_type(key) == filter_type)

        def process_pat(keys: Iterable[str]) -> Iterable[str]:
            if pat is None:
                return keys
            return (
                key
                for key in keys
                if pat.match(key))

        def process(keys: list[str]) -> list[str]:
            if filter_type is None and pat is None:
                return keys
            return list(process_pat(process_type(keys)))

        next_cursor, keys = self._scan(cursor, count, prefix, now_mono)
        if prefix is not None and keys:
            last_key = keys[-1]
            last_key = last_key[:len(prefix)]
            if last_key > prefix:
                next_cursor = 0
        return next_cursor, process(keys)

    def get_all_keys(
            self,
            now_mono: float,
            *,
            match: str | None,
            filter_type: KeyType | None) -> Iterable[str]:
        """
        Completes a full scan through all keys. As this operation is performed
        fully inside a lock, no duplicate keys will be returned.

        Args:
            now_mono (float): The current time.
            match (str | None): If not None a redis key pattern to filter
                keys.
            filter_type (KeyType | None): If not None filter by key type.

        Yields:
            str: The key.
        """
        count = MIN_SCAN_LENGTH
        cursor = 0
        while True:
            cursor, keys = self.scan_keys(
                cursor,
                now_mono,
                count=count,
                match=match,
                filter_type=filter_type)
            yield from keys
            if cursor == 0:
                break
            count = min(count * 2, KEYS_MAX_SCAN)

    def exists(self, key: str) -> bool:
        """
        Checks whether a given key exists. This method does not account for
        expiration.

        Args:
            key (str): The key.

        Returns:
            bool: Whether it exists as any type.
        """
        if self._is_pending_delete(key):
            return False
        if key in self._vals:
            return True
        if key in self._queues:
            return True
        if key in self._hashes:
            return True
        if key in self._sets:
            return True
        if key in self._zorder:
            return True
        if self._parent is not None:
            return self._parent.exists(key)
        return False

    def copy_from_parent(self, key: str) -> None:
        """
        Copies the current value from the parent if the value of the key is not
        already present in the own dictionary.

        Args:
            key (str): The key.
        """
        if self._is_pending_delete(key):
            return
        parent = self._parent
        if parent is None:
            return
        parent.copy_from_parent(key)

        def copy_over(from_dict: dict[str, T], to_dict: dict[str, T]) -> None:
            if key in to_dict:
                return
            other = from_dict.get(key)
            if other is not None:
                to_dict[key] = other

        copy_over(parent.raw_vals(), self._vals)
        copy_over(parent.raw_queues(), self._queues)
        copy_over(parent.raw_hashes(), self._hashes)
        copy_over(parent.raw_sets(), self._sets)
        copy_over(parent.raw_zorder(), self._zorder)
        copy_over(parent.raw_zscores(), self._zscores)

    def delete(self, deletes: set[str]) -> None:
        """
        Deletes keys.

        Args:
            deletes (set[str]): The keys to delete.
        """
        has_delete = False
        if self._parent is not None:
            self._deletes.update(deletes)
            has_delete = True
        for key in deletes:
            if self._expire.pop(key, None) is not None:
                has_delete = True
            if self._vals.pop(key, None) is not None:
                has_delete = True
            if self._queues.pop(key, None) is not None:
                has_delete = True
            if self._hashes.pop(key, None) is not None:
                has_delete = True
            if self._sets.pop(key, None) is not None:
                has_delete = True
            if self._zorder.pop(key, None) is not None:
                has_delete = True
            if self._zscores.pop(key, None) is not None:
                has_delete = True
        if has_delete:
            self._flush_key_cache()
            self._delete_count += 1

    def apply(self, other: 'State', now_mono: float) -> None:
        """
        Applies the given state. All updates that are different in the other
        state will be updated in the current state and all deleted keys in
        the other state will delete corresponding keys in the current state.
        If a key in the other state has a different type than the key in the
        current state the type in the current state will be changed. This
        relies on the invariant that a key can only appear once in the other
        and once in the current state.

        Args:
            other (State): The state to be applied to the current state.
            now_mono (float): The current time.
        """
        raw_expire = other.raw_expirations()
        raw_vals = other.raw_vals()
        raw_queues = other.raw_queues()
        raw_hashes = other.raw_hashes()
        raw_sets = other.raw_sets()
        raw_zorder = other.raw_zorder()
        raw_zscores = other.raw_zscores()

        new_keys: set[str] = set()  # NOTE: new means new for the given type
        new_keys.update(set(raw_expire.keys()) - set(self._expire.keys()))
        new_keys.update(set(raw_vals.keys()) - set(self._vals.keys()))
        new_keys.update(set(raw_queues.keys()) - set(self._queues.keys()))
        new_keys.update(set(raw_hashes.keys()) - set(self._hashes.keys()))
        new_keys.update(set(raw_sets.keys()) - set(self._sets.keys()))
        new_keys.update(set(raw_zorder.keys()) - set(self._zorder.keys()))
        new_keys.update(set(raw_zscores.keys()) - set(self._zscores.keys()))
        # NOTE: deleting all new keys makes sure they don't exist as a
        # different type
        self.delete(new_keys)

        self._expire.update(raw_expire)
        self._vals.update(raw_vals)
        self._queues.update(raw_queues)
        self._hashes.update(raw_hashes)
        self._sets.update(raw_sets)
        self._zorder.update(raw_zorder)
        self._zscores.update(raw_zscores)
        self.clean_vals(now_mono)
        self.delete(other.raw_deletes())

        if new_keys:
            self._flush_key_cache()

    def reset(self) -> None:
        """
        Completely resets the state.
        """
        self._expire.clear()
        self._vals.clear()
        self._queues.clear()
        self._hashes.clear()
        self._sets.clear()
        self._zorder.clear()
        self._zscores.clear()
        self._deletes.clear()
        self._flush_key_cache()

    def _prepare_key_for_write(self, key: str, *, is_new: bool) -> None:
        """
        Prepares a key for writing by ensuring that it is not deleted.

        Args:
            key (str): The key.
            is_new (bool): Whether the key is new.
        """
        exists = self.exists(key)
        if not exists or is_new:
            self._flush_key_cache()
        if exists:
            return
        self._deletes.discard(key)

    def _is_pending_delete(self, key: str) -> bool:
        """
        Whether the given key is marked as deleted. Note, this is not the case
        after _prepare_key_for_write.

        Args:
            key (str): The key.

        Returns:
            bool: Whether the given key is in the delete set.
        """
        return key in self._deletes

    def is_alive(self, key: str, now_mono: float) -> bool:
        """
        Checks whether the given key has not expired.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            bool: Whether the key is alive.
        """
        expire = self.get_expire(key)
        if expire is None:
            return True
        if expire > now_mono:
            return True
        return False

    def clean_vals(self, now_mono: float) -> None:
        """
        Cleans up expired values by deleting them.

        Args:
            now_mono (float): The current time.
        """
        if self._parent is not None:
            return
        remove = set()
        for key, value in self._vals.items():
            if self.is_alive(value, now_mono):
                continue
            remove.add(key)
        self.delete(remove)

    def raw_expirations(self) -> dict[str, float]:
        """
        Raw access to the expiration dictionary.

        Returns:
            dict[str, float]: The actual expiration dictionary of this state.
                Here persistent is represented as a value <= 0.
        """
        return self._expire

    def raw_vals(self) -> dict[str, str]:
        """
        Raw access to the value dictionary.

        Returns:
            dict[str, str]: The actual value dictionary of this state.
        """
        return self._vals

    def raw_queues(
            self) -> dict[str, collections.deque[str]]:
        """
        Raw access to the queues dictionary.

        Returns:
            dict[str, collections.deque[str]]: The actual queue dictionary of
                this state.
        """
        return self._queues

    def raw_hashes(self) -> dict[str, dict[str, str]]:
        """
        Raw access to the hashes dictionary.

        Returns:
            dict[str, dict[str, str]]: The actual hashes dictionary of this
                state.
        """
        return self._hashes

    def raw_sets(self) -> dict[str, set[str]]:
        """
        Raw access to the sets dictionary.

        Returns:
            dict[str, set[str]]: The actual sets dictionary of this state.
        """
        return self._sets

    def raw_zorder(self) -> dict[str, list[str]]:
        """
        Raw access to the zorder dictionary.

        Returns:
            dict[str, list[str]]: The actual zorder dictionary of this state.
        """
        return self._zorder

    def raw_zscores(self) -> dict[str, dict[str, float]]:
        """
        Raw access to the zscores dictionary.

        Returns:
            dict[str, dict[str, float]]: The actual zscores dictionary of this
                state.
        """
        return self._zscores

    def raw_deletes(self) -> set[str]:
        """
        Raw access to the deletes set.

        Returns:
            set[str]: The actual deletes set of this state.
        """
        return self._deletes

    def set_value(self, key: str, value: str, now_mono: float) -> None:
        """
        Sets the value for the given key.

        Args:
            key (str): The key.
            value (str): The value.
            now_mono (float): The current time.
        """
        if not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
        is_new = False
        if key not in self._vals:
            self.verify_key("string", key)
            is_new = True
        self._prepare_key_for_write(key, is_new=is_new)
        self._vals[key] = value

    def get_value(
            self,
            key: str,
            now_mono: float) -> str | None:
        """
        Retrieves the value associated with the given key.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            tuple[str, float | None] | None: The value of the key and when it
                will expire.
        """
        res = self._vals.get(key)
        if res is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            return None
        if res is None:
            self.verify_key("string", key)
            if self._parent is not None and not self._is_pending_delete(key):
                return self._parent.get_value(key, now_mono)
        return res

    def get_expire(self, key: str) -> float | None:
        """
        Retrieves the expiration of the given key.

        Args:
            key (str): The key.

        Returns:
            float | None: The raw expiration.
        """
        res = self._expire.get(key)
        if res is None:
            if self._parent is not None and not self._is_pending_delete(key):
                return self._parent.get_expire(key)
        if res is not None and res <= 0.0:
            res = None
        return res

    def expire(
            self,
            key: str,
            expire_fn: Callable[[float | None], float | None]) -> bool:
        """
        Conditionally sets the expiration for the given key.

        Args:
            key (str): The key.
            expire_fn (Callable[[float | None], float | None]): Function
                to determine the new expiration value. The argument is the
                previous expiration value.

        Returns:
            bool: Whether the expiration changed.
        """
        if not self.exists(key):
            return False
        prev_expire = self.get_expire(key)
        new_expire = expire_fn(prev_expire)
        if new_expire == prev_expire:
            return False
        self.copy_from_parent(key)
        self._expire[key] = -1.0 if new_expire is None else new_expire
        return True

    def ttl(self, key: str, now_mono: float) -> float | None:
        """
        Computes the time-to-live for the given key.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            float | None: The time in seconds until the key expires. If the
                value is `<= 0.0` it means the key does not have an expiration.
                If the result is None the key does not exist.
        """
        expire = self.get_expire(key)
        if expire is None:
            if not self.exists(key):
                return None
            return -1.0
        res_expire = expire - now_mono
        if res_expire <= 0.0:
            self.clean_vals(now_mono)
            return None
        return res_expire

    def get_queue(self, key: str, now_mono: float) -> collections.deque[str]:
        """
        Returns a queue for writing. If the queue doesn't exist already it will
        be created.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            collections.deque[str]: The queue for writing.
        """
        res = self._queues.get(key)
        if res is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            res = None
        is_new = False
        if res is None:
            self.verify_key("list", key)
            if self._parent is not None and not self._is_pending_delete(key):
                res = collections.deque(self._parent.get_queue(key, now_mono))
            else:
                res = collections.deque()
            self._queues[key] = res
            is_new = True
        self._prepare_key_for_write(key, is_new=is_new)
        return res

    def readonly_queue(
            self, key: str, now_mono: float) -> collections.deque[str] | None:
        """
        Returns a queue for reading only. If the queue doesn't exist None is
        returned.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            collections.deque[str] | None: The readonly queue. Make sure to not
                modify it as it is not a copy. None if the queue doesn't exist.
        """
        res = self._queues.get(key)
        if res is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            return None
        if res is None:
            self.verify_key("list", key)
            if self._parent is None or self._is_pending_delete(key):
                return None
            return self._parent.readonly_queue(key, now_mono)
        return res

    def queue_len(self, key: str, now_mono: float) -> int:
        """
        Computes the length of a queue.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            int: The length of the queue.
        """
        res = self.readonly_queue(key, now_mono)
        if res is None:
            self.verify_key("list", key)
            return 0
        return len(res)

    def get_hash(self, key: str, now_mono: float) -> dict[str, str]:
        """
        Returns a hash for writing. If the hash doesn't exist already it will
        be created.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            dict[str, str]: The hash for writing.
        """
        res = self._hashes.get(key)
        if res is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            res = None
        is_new = False
        if res is None:
            self.verify_key("hash", key)
            if self._parent is not None and not self._is_pending_delete(key):
                res = dict(self._parent.get_hash(key, now_mono))
            else:
                res = {}
            self._hashes[key] = res
            is_new = True
        self._prepare_key_for_write(key, is_new=is_new)
        return res

    def readonly_hash(
            self, key: str, now_mono: float) -> dict[str, str] | None:
        """
        Returns a hash for reading only. If the hash doesn't exist None is
        returned.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            dict[str, str] | None: The readonly hash. Make sure to not
                modify it as it is not a copy. None if the hash doesn't exist.
        """
        res = self._hashes.get(key)
        if res is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            return None
        if res is None:
            self.verify_key("hash", key)
            if self._parent is None or self._is_pending_delete(key):
                return None
            return self._parent.readonly_hash(key, now_mono)
        return res

    def get_set(self, key: str, now_mono: float) -> set[str]:
        """
        Returns a set for writing. If the set doesn't exist already it will
        be created.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            set[str]: The set for writing.
        """
        res = self._sets.get(key)
        if res is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            res = None
        is_new = False
        if res is None:
            self.verify_key("set", key)
            if self._parent is not None and not self._is_pending_delete(key):
                res = set(self._parent.get_set(key, now_mono))
            else:
                res = set()
            self._sets[key] = res
            is_new = True
        self._prepare_key_for_write(key, is_new=is_new)
        return res

    def readonly_set(self, key: str, now_mono: float) -> set[str] | None:
        """
        Returns a set for reading only. If the set doesn't exist None is
        returned.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            set[str] | None: The readonly set. Make sure to not
            modify it as it is not a copy. None if the set doesn't exist.
        """
        res = self._sets.get(key)
        if res is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            return None
        if res is None:
            self.verify_key("set", key)
            if self._parent is None or self._is_pending_delete(key):
                return None
            return self._parent.readonly_set(key, now_mono)
        return res

    def get_zset(
            self,
            key: str,
            now_mono: float) -> tuple[list[str], dict[str, float]]:
        """
        Returns a zset for writing. If the zset doesn't exist already it
        will be created.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            tuple[list[str], dict[str, float]]: The tuple of zorder and zscores
            for writing.
        """
        rorder = self._zorder.get(key)
        rscores = self._zscores.get(key)
        if rorder is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            rorder = None
            rscores = None
        is_new = False
        if rorder is None:
            self.verify_key("zset", key)
            if self._parent is not None and not self._is_pending_delete(key):
                porder, pscores = self._parent.get_zset(key, now_mono)
                rorder = list(porder)
                rscores = dict(pscores)
            else:
                rorder = []
                rscores = {}
            self._zorder[key] = rorder
            self._zscores[key] = rscores
            is_new = True
        self._prepare_key_for_write(key, is_new=is_new)
        assert rscores is not None
        return (rorder, rscores)

    def readonly_zorder(self, key: str, now_mono: float) -> list[str] | None:
        """
        Returns a zorder for reading only. If the zorder doesn't exist None is
        returned.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            list[str] | None: The readonly zorder. Make sure to not
            modify it as it is not a copy. None if the zorder doesn't exist.
        """
        res = self._zorder.get(key)
        if res is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            return None
        if res is None:
            self.verify_key("zset", key)
            if self._parent is None or self._is_pending_delete(key):
                return None
            return self._parent.readonly_zorder(key, now_mono)
        return res

    def readonly_zscores(
            self, key: str, now_mono: float) -> dict[str, float] | None:
        """
        Returns a zscores for reading only. If the zscores doesn't exist None
        is returned.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            dict[str, float] | None: The readonly zscores. Make sure to not
            modify it as it is not a copy. None if the zscores doesn't exist.
        """
        res = self._zscores.get(key)
        if res is not None and not self.is_alive(key, now_mono):
            self.clean_vals(now_mono)
            return None
        if res is None:
            self.verify_key("zset", key)
            if self._parent is None or self._is_pending_delete(key):
                return None
            return self._parent.readonly_zscores(key, now_mono)
        return res

    def zorder_len(self, key: str, now_mono: float) -> int:
        """
        Computes the length of a zorder.

        Args:
            key (str): The key.
            now_mono (float): The current time.

        Returns:
            int: The length of the zorder.
        """
        res = self.readonly_zorder(key, now_mono)
        if res is None:
            self.verify_key("zset", key)
            return 0
        return len(res)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}["
            f"expire={self._expire},"
            f"vals={self._vals},"
            f"queues={dict(self._queues)},"
            f"hashes={dict(self._hashes)},"
            f"sets={dict(self._sets)},"
            f"zorder={dict(self._zorder)},"
            f"zscores={dict(self._zscores)},"
            f"deletes={self._deletes},"
            f"has_parent={self._parent is not None}]")

    def __repr__(self) -> str:
        return self.__str__()


class Machine(RedisAPI):
    """
    A Machine manages the state of a memory runtime and exposes the redis API.
    """
    def __init__(self, state: State, plock: threading.RLock) -> None:
        """
        Creates a Machine.

        Args:
            state (State): The associated state.
            plock (RLock): The pubsub lock.
        """
        super().__init__()
        self._state = state
        self._now_mono: tuple[float, datetime] | None = None
        # FIXME: for now we implement pubsub in the machine
        self._pubsub: dict[str, tuple[threading.Condition, int]] = {}
        self._plock = plock

    def plock(self) -> threading.RLock:
        """
        The pubsub lock.

        Returns:
            threading.RLock: The pubsub lock.
        """
        return self._plock

    def set_mono(self, now_mono: tuple[float, datetime] | None) -> None:
        """
        Sets the current time.

        Args:
            now_mono (tuple[float, datetime] | None): The current time for
                pipelines as monotonic time and datetime. Otherwise None.
        """
        self._now_mono = now_mono

    def get_mono(self) -> float:
        """
        Returns the current monotonic time.

        Returns:
            float: The current time or the time associated with this machine if
                it is a pipeline.
        """
        if self._now_mono is None:
            return time.monotonic()
        return self._now_mono[0]

    def get_ts(self) -> datetime:
        """
        Returns the current time.

        Returns:
            datetime: The current time or the time associated with this machine
                if it is a pipeline.
        """
        return now() if self._now_mono is None else self._now_mono[1]

    def get_state(self) -> State:
        """
        Returns the managed state.

        Returns:
            State: The associated state.
        """
        return self._state

    def exists(self, *keys: str) -> int:
        now_mono = self.get_mono()
        res = 0
        for key in keys:
            if self._state.exists(key) and self._state.is_alive(key, now_mono):
                res += 1
        return res

    def delete(self, *keys: str) -> int:
        res = self.exists(*keys)
        self._state.delete(set(keys))
        return res

    def key_type(self, key: str) -> KeyType | None:
        now_mono = self.get_mono()
        if not self._state.is_alive(key, now_mono):
            return None
        return self._state.key_type(key)

    def scan(
            self,
            cursor: int,
            *,
            match: str | None = None,
            count: int | None = None,
            filter_type: KeyType | None = None) -> tuple[int, list[str]]:
        now_mono = self.get_mono()
        return self._state.scan_keys(
            cursor,
            now_mono,
            count=10 if count is None else count,
            match=match,
            filter_type=filter_type)

    def keys_block(
            self,
            *,
            match: str | None = None,
            filter_type: KeyType | None = None) -> list[str]:
        now_mono = self.get_mono()
        # NOTE: get_all_keys cannot have duplicates
        return list(self._state.get_all_keys(
            now_mono, match=match, filter_type=filter_type))

    def flushall(self) -> None:
        # NOTE: this method cannot be used in a pipeline as the effect
        # is immediate!
        if self._state.has_parent():
            raise RuntimeError("cannot use flushall with a parent!")
        self._state.reset()

    @overload
    def set_value(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[True],
            expire_timestamp: datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> str | None:
        ...

    @overload
    def set_value(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[False],
            expire_timestamp: datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> bool | None:
        ...

    @overload
    def set_value(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        ...

    def set_value(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        state = self._state
        now_mono = self.get_mono()
        now_ts = self.get_ts()
        expire = compute_expire(
            now_mono,
            now_ts,
            expire_timestamp=expire_timestamp,
            expire_in=expire_in)
        prev_value = state.get_value(key, now_mono)
        do_set = False
        if mode == RSM_ALWAYS:
            do_set = True
        elif mode == RSM_EXISTS:
            do_set = prev_value is not None
        elif mode == RSM_MISSING:
            do_set = prev_value is None
        else:
            raise ValueError(f"unknown mode: {mode}")
        if do_set:
            state.set_value(key, value, now_mono)
            state.expire(
                key, lambda prev_expire: prev_expire if keep_ttl else expire)
        if return_previous:
            return prev_value
        return do_set

    def get_value(self, key: str) -> str | None:
        now_mono = self.get_mono()
        return self._state.get_value(key, now_mono)

    def expire(
            self,
            key: str,
            *,
            mode: RExpireMode = REX_ALWAYS,
            expire_timestamp: datetime | None = None,
            expire_in: float | None = None) -> bool:
        state = self._state
        now_mono = self.get_mono()
        now_ts = self.get_ts()
        expire = compute_expire(
            now_mono,
            now_ts,
            expire_timestamp=expire_timestamp,
            expire_in=expire_in)

        def choose_expire(prev_expire: float | None) -> float | None:
            if expire is None:
                return prev_expire if mode == REX_EARLIER else None
            if mode == REX_ALWAYS:
                return expire
            if mode == REX_PERSIST:
                return expire if prev_expire is None else prev_expire
            if mode == REX_EXPIRE:
                return None if prev_expire is None else expire
            if mode == REX_EARLIER:
                if prev_expire is None:
                    return expire
                return min(expire, prev_expire)
            if mode == REX_LATER:
                if prev_expire is None:
                    return None
                return max(expire, prev_expire)
            raise ValueError(f"unknown mode: {mode}")

        if not state.is_alive(key, now_mono) or not state.exists(key):
            return False
        return state.expire(key, choose_expire)

    def ttl(self, key: str) -> float | None:
        now_mono = self.get_mono()
        return self._state.ttl(key, now_mono)

    def incrby(self, key: str, inc: float | int) -> float:
        now_mono = self.get_mono()
        val = self._state.get_value(key, now_mono)
        if val is None:
            num = inc
        else:
            num = float(val) + inc
        self._state.set_value(key, to_number_str(num), now_mono)
        return num

    def lpush(self, key: str, *values: str) -> int:
        now_mono = self.get_mono()
        queue = self._state.get_queue(key, now_mono)
        queue.extendleft(values)
        return len(queue)

    def rpush(self, key: str, *values: str) -> int:
        now_mono = self.get_mono()
        queue = self._state.get_queue(key, now_mono)
        queue.extend(values)
        return len(queue)

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        now_mono = self.get_mono()
        queue = self._state.get_queue(key, now_mono)
        if not queue:
            return None
        if count is None:
            rval = queue.popleft()
            if not queue:
                self.delete(key)
            return rval
        popc = count
        res = []
        while popc > 0 and queue:
            res.append(queue.popleft())
            popc -= 1
        if not queue:
            self.delete(key)
        return res if res else None

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        now_mono = self.get_mono()
        queue = self._state.get_queue(key, now_mono)
        if not queue:
            return None
        if count is None:
            rval = queue.pop()
            if not queue:
                self.delete(key)
            return rval
        popc = count
        res = []
        while popc > 0 and queue:
            res.append(queue.pop())
            popc -= 1
        if not queue:
            self.delete(key)
        return res if res else None

    def lrange(self, key: str, start: int, stop: int) -> list[str]:
        now_mono = self.get_mono()
        queue = self._state.readonly_queue(key, now_mono)
        if queue is None:
            return []
        if start >= len(queue):
            return []
        if start < 0:
            start = max(0, start + len(queue))
        if stop < 0:
            stop += len(queue)
            if stop < 0:
                return []
        stop += 1
        queue.rotate(-start)
        res = list(itertools.islice(queue, 0, stop - start, 1))
        queue.rotate(start)
        return res

    def lset(self, key: str, index: int, value: str) -> None:
        now_mono = self.get_mono()
        queue = self._state.get_queue(key, now_mono)
        queue[index] = value

    def lindex(self, key: str, index: int) -> str | None:
        now_mono = self.get_mono()
        queue = self._state.readonly_queue(key, now_mono)
        if not queue:
            return None
        try:
            return queue[index]
        except IndexError:
            return None

    def llen(self, key: str) -> int:
        now_mono = self.get_mono()
        return self._state.queue_len(key, now_mono)

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        now_mono = self.get_mono()
        count = 0
        zorder, zscores = self._state.get_zset(key, now_mono)
        for name, score in mapping.items():
            if name not in zscores:
                zorder.append(name)
                count += 1
            zscores[name] = score
        zorder.sort(key=lambda k: (zscores[k], k))
        return count

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        now_mono = self.get_mono()
        zorder, zscores = self._state.get_zset(key, now_mono)
        res = []
        remain = 1 if count is None else count
        while remain > 0 and zorder:
            name = zorder.pop()
            score = zscores.pop(name)
            res.append((name, score))
            remain -= 1
        return res

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        now_mono = self.get_mono()
        zorder, zscores = self._state.get_zset(key, now_mono)
        res = []
        remain = 1 if count is None else count
        while remain > 0 and zorder:
            name = zorder.pop(0)
            score = zscores.pop(name)
            res.append((name, score))
            remain -= 1
        return res

    def zrange(self, key: str, start: int, stop: int) -> list[str]:
        now_mono = self.get_mono()
        zorder = self._state.readonly_zorder(key, now_mono)
        if zorder is None:
            return []
        astop: int | None = stop
        if astop == -1 or astop is None:  # NOTE: mypy workaround
            astop = None
        else:
            astop += 1
        return zorder[start:astop]

    def zcard(self, key: str) -> int:
        now_mono = self.get_mono()
        return self._state.zorder_len(key, now_mono)

    def hset(self, key: str, mapping: dict[str, str]) -> int:
        now_mono = self.get_mono()
        obj = self._state.get_hash(key, now_mono)
        res = len(set(mapping.keys()).difference(obj.keys()))
        obj.update(mapping)
        return res

    def hdel(self, key: str, *fields: str) -> int:
        now_mono = self.get_mono()
        obj = self._state.get_hash(key, now_mono)
        res = 0
        for field in fields:
            if obj.pop(field, None) is not None:
                res += 1
        if not obj:
            self._state.delete({key})
        return res

    def hget(self, key: str, field: str) -> str | None:
        now_mono = self.get_mono()
        res = self._state.readonly_hash(key, now_mono)
        if res is None:
            return None
        return res.get(field)

    def hmget(self, key: str, *fields: str) -> dict[str, str | None]:
        now_mono = self.get_mono()
        res = self._state.readonly_hash(key, now_mono)
        if res is None:
            return {}
        return {
            field: res.get(field)
            for field in fields
        }

    def hincrby(self, key: str, field: str, inc: float | int) -> float:
        now_mono = self.get_mono()
        res = self._state.get_hash(key, now_mono)
        num = float(res.get(field, 0)) + inc
        res[field] = to_number_str(num)
        return num

    def hkeys(self, key: str) -> list[str]:
        now_mono = self.get_mono()
        res = self._state.readonly_hash(key, now_mono)
        if res is None:
            return []
        return list(res.keys())

    def hvals(self, key: str) -> list[str]:
        now_mono = self.get_mono()
        res = self._state.readonly_hash(key, now_mono)
        if res is None:
            return []
        return list(res.values())

    def hgetall(self, key: str) -> dict[str, str]:
        now_mono = self.get_mono()
        res = self._state.readonly_hash(key, now_mono)
        if res is None:
            return {}
        return dict(res)

    def sadd(self, key: str, *values: str) -> int:
        now_mono = self.get_mono()
        obj = self._state.get_set(key, now_mono)
        before = len(obj)
        obj.update(values)
        return len(obj) - before

    def srem(self, key: str, *values: str) -> int:
        now_mono = self.get_mono()
        obj = self._state.get_set(key, now_mono)
        before = len(obj)
        obj.difference_update(set(values))
        if not obj:
            self._state.delete({key})
        return before - len(obj)

    def sismember(self, key: str, value: str) -> bool:
        now_mono = self.get_mono()
        obj = self._state.readonly_set(key, now_mono)
        if obj is None:
            return False
        return value in obj

    def scard(self, key: str) -> int:
        now_mono = self.get_mono()
        obj = self._state.readonly_set(key, now_mono)
        if obj is None:
            return 0
        return len(obj)

    def smembers(self, key: str) -> set[str]:
        now_mono = self.get_mono()
        obj = self._state.readonly_set(key, now_mono)
        if obj is None:
            return set()
        return set(obj)

    def publish(self, key: str, msg: str) -> None:
        # FIXME: we do not support patterns or messages yet
        pubsub_key = reject_patterns(key)
        with self.plock():
            channel = self._pubsub.get(pubsub_key)
            if channel is None:
                return
            listen, _ = channel
            listen.notify_all()

    def wait_for(
            self,
            key: str,
            predicate: Callable[[], T],
            timeout: float | None) -> T | None:
        # FIXME: we do not support patterns or messages yet
        pubsub_key = reject_patterns(key)
        plock = self.plock()
        with plock:
            channel = self._pubsub.get(pubsub_key)
            try:
                if channel is None:
                    listen = threading.Condition(plock)
                    before_count = 0
                else:
                    listen, before_count = channel
                self._pubsub[pubsub_key] = (listen, before_count + 1)
                res = listen.wait_for(predicate, timeout)
            finally:
                channel = self._pubsub.get(pubsub_key)
                if channel is not None:
                    listen, after_count = channel
                    if after_count > 1:
                        self._pubsub[pubsub_key] = (listen, after_count - 1)
                    else:
                        self._pubsub.pop(pubsub_key, None)
        return res

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}[state={self._state}]")

    def __repr__(self) -> str:
        return self.__str__()



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines the basic redis API. All redis functions appear once
in RedisAPI and once in PipelineAPI. Additional functionality is added via
RedisClientAPI."""
import contextlib
import datetime
from collections.abc import Callable, Iterable, Iterator
from typing import cast, get_args, Literal, overload, TypeVar

from redipy.backend.backend import ExecFunction
from redipy.symbolic.seq import FnContext


T = TypeVar('T')


RSetMode = Literal[
    "always",
    "if_missing",  # NX
    "if_exists",  # XX
]
"""The conditions on when to set a value for the set command."""
RSM_ALWAYS: RSetMode = "always"
"""The value will always be set."""
RSM_MISSING: RSetMode = "if_missing"
"""The value will only be set when the key was missing.
This is equivalent to the NX flag."""
RSM_EXISTS: RSetMode = "if_exists"
"""The value will only be set when the key did exist.
This is equivalent to the XX flag."""


RExpireMode = Literal[
    "always",
    "if_persist",  # NX
    "if_expire",  # XX
    "if_later",  # GT
    "if_earlier",  # LT
]
"""The conditions on when to expire a value for the expire command."""
REX_ALWAYS: RExpireMode = "always"
"""The expiration will always be set."""
REX_PERSIST: RExpireMode = "if_persist"
"""The expiration will be set only if there is no previous expiration.
This is equivalent to the NX flag."""
REX_EXPIRE: RExpireMode = "if_expire"
"""The expiration will be set only if there is a previous expiration.
This is equivalent to the XX flag."""
REX_LATER: RExpireMode = "if_later"
"""The expiration will be set only if it is later than the previous expiration.
This is equivalent to the GT flag."""
REX_EARLIER: RExpireMode = "if_earlier"
"""The expiration will be set only if it is earlier than the previous
expiration. This is equivalent to the LT flag."""


KeyType = Literal[
    "string",
    "list",
    "set",
    "zset",
    "hash",
    "stream",
]
"""The different key types."""


KEY_TYPES: set[KeyType] = set(get_args(KeyType))


@overload
def as_key_type(text: str) -> KeyType | None:
    ...


@overload
def as_key_type(text: None) -> None:
    ...


def as_key_type(text: str | None) -> KeyType | None:
    """
    Converts a string into a key type.

    Args:
        text (str): The string.

    Raises:
        ValueError: If the string does not represent a key type.

    Returns:
        KeyType: The key type or None if the input was None or the input was
            the string "none".
    """
    if text is None or text == "none":
        return None
    if text not in KEY_TYPES:
        raise ValueError(
            f"unknown key type: {text}. Only {KEY_TYPES} are supported.")
    return cast(KeyType, text)


class PipelineAPI:
    """Redis API as pipeline. All methods return None and you have to call
    execute to retrieve the results of the pipeline commands."""
    def execute(self) -> list:
        """
        Executes the pipeline and returns the result values of each command.

        Returns:
            list: The result values of each command.
        """
        raise NotImplementedError()

    def exists(self, *keys: str) -> None:
        """
        Determines whether specified keys exist.

        See also the redis documentation: https://redis.io/commands/exists/

        The pipeline value is set to the number of keys that exist.

        Args:
            *keys (str): The keys.
        """
        raise NotImplementedError()

    def delete(self, *keys: str) -> None:
        """
        Deletes keys.

        See also the redis documentation: https://redis.io/commands/del/

        The pipeline value is set to the number of keys that got removed.

        Args:
            *keys (str): The keys.
        """
        raise NotImplementedError()

    def key_type(self, key: str) -> None:
        """
        The type of the given key if it exists.

        See also the redis documentation: https://redis.io/commands/type/

        The pipeline value is the key type or None if the key does not exist.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def scan(
            self,
            cursor: int,
            *,
            match: str | None = None,
            count: int | None = None,
            filter_type: KeyType | None = None) -> None:
        """
        Performs a singular scan operation on the keys of the database. Note,
        that this cannot be used to scan all keys in a pipeline as scan depends
        on the previous' scan cursor which is unobtainable within the same
        pipeline.

        See also the redis documentation: https://redis.io/commands/scan/

        A tuple of the new cursor and the current keys. If the new cursor is
        0 the iteration ends.

        Args:
            cursor (int): The cursor. This value is either 0 (for starting a
                scan) or a value returned previously by this function.
            match (str | None, optional): Filters the keys according to a redis
                match string. Defaults to None.
            count (int | None, optional): Estimate of expected returned keys
                in one call. The actual number returned might be different.
                Defaults to None.
            filter_type (KeyType | None, optional): Filters by the key type.
                Defaults to None.
        """
        raise NotImplementedError()

    def keys(
            self,
            *,
            match: str | None = None,
            filter_type: KeyType | None = None) -> None:
        """
        Retrieves all matching keys. This method can take very long to complete
        and block the database for a long time. It is best to avoid retrieving
        keys in a pipeline.

        See also the redis documentation: https://redis.io/commands/keys/

        The pipeline value is the list of unique matching keys.

        Args:
            match (str | None, optional): Filters the keys according to a redis
                match string. Defaults to None.
            filter_type (KeyType | None, optional): Filters by the key type.
                Defaults to None.
        """
        raise NotImplementedError()

    def set_value(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> None:
        """
        Sets a value for a given key. The value can be scheduled to expire.

        See also the redis documentation: https://redis.io/commands/set/

        The pipeline value depends on the return_previous argument.

        Args:
            key (str): The key.

            value (str): The value.

            mode (RSetMode, optional): Under which condition to set the value
                valid values are RSM_ALWAYS, RSM_MISSING, and RSM_EXISTS.
                RSM_MISSING is the equivalent of setting the NX flag.
                RSM_EXISTS is the equivalent of the XX flag. Defaults to
                RSM_ALWAYS.

            return_previous (bool, optional): Whether to return the previous
                value associated with the key. Defaults to False.

            expire_timestamp (datetime.datetime | None, optional): A timestamp
                on when to expire the key. Defaults to None.

            expire_in (float | None, optional): A relative time in seconds on
                when to expire the key. Defaults to None.

            keep_ttl (bool, optional): Whether to keep previous expiration
                times. Defaults to False.
        """
        raise NotImplementedError()

    def get_value(self, key: str) -> None:
        """
        Retrieves the value for the given key.

        See also the redis documentation: https://redis.io/commands/get/

        The pipeline value is the value or None if the key does not exists or
        the value has expired.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def expire(
            self,
            key: str,
            *,
            mode: RExpireMode = REX_ALWAYS,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None) -> None:
        """
        Sets or removes the expiration time of a key. The expiration can either
        be a timestamp or a relative amount. If no time is set the expiration
        is removed and the key persists.

        See also the redis documentation:
        https://redis.io/commands/expire/
        and https://redis.io/commands/expireat/
        and https://redis.io/commands/expiretime/
        and https://redis.io/commands/persist/
        and https://redis.io/commands/pexpire/
        and https://redis.io/commands/pexpireat/
        and https://redis.io/commands/pexpiretime/

        The pipeline value is set to whether a change in expiration occurred.

        Args:
            key (str): The key.

            mode (RExpireMode, optional): The expiration mode. Defaults to
                REX_ALWAYS. The mode is ignored if all expirations are None.

            expire_timestamp (datetime.datetime | None, optional): A timestamp
                on when to expire the key. Defaults to None.

            expire_in (float | None, optional): A relative time in seconds on
                when to expire the key. Defaults to None.
        """
        raise NotImplementedError()

    def ttl(self, key: str) -> None:
        """
        Retrieves the time-to-live for a key.

        See also the redis documentation:
        https://redis.io/commands/ttl/ and https://redis.io/commands/pttl/

        The pipeline value is set to the time to live in seconds with
        millisecond precision. If the key is persistent the time is `<= 0.0`.
        If the key does not exist None is returned.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def incrby(self, key: str, inc: float | int) -> None:
        """
        Updates the value associated with the given key by a relative amount.
        The value is interpreted as number. If the value doesn't exist zero is
        used as starting point.

        See also the redis documentation:
        https://redis.io/commands/incrby/
        https://redis.io/commands/incrbyfloat/

        The pipeline value is set to the new value as float.
        If the value cannot be interpreted as float while executing the
        pipeline a ValueError exception is raised.

        Args:
            key (str): The key.

            inc (float | int): The relative change.
        """
        raise NotImplementedError()

    def lpush(self, key: str, *values: str) -> None:
        """
        Pushes values to the left side of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/lpush/

        The pipeline value is the length of the list after the push.

        Args:
            key (str): The key.

            *values (str): The values to push.
        """
        raise NotImplementedError()

    def rpush(self, key: str, *values: str) -> None:
        """
        Pushes values to the right side of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/rpush/

        The pipeline value is the length of the list after the push.

        Args:
            key (str): The key.

            *values (str): The values to push.
        """
        raise NotImplementedError()

    def lpop(
            self,
            key: str,
            count: int | None = None) -> None:
        """
        Pops a number of values from the left side of the list associated with
        the key.

        See also the redis documentation: https://redis.io/commands/lpop/

        The pipeline value is None if the key doesn't exist. If a count
        is set a list with values in pop order is set as pipeline value (even
        if it is set to one). If count is not set (default or None) the single
        value that got popped is set as pipeline value.

        Args:
            key (str): The key.

            count (int | None, optional): The number values to pop.
            Defaults to a single value.
        """
        raise NotImplementedError()

    def rpop(
            self,
            key: str,
            count: int | None = None) -> None:
        """
        Pops a number of values from the right side of the list associated with
        the key.

        See also the redis documentation: https://redis.io/commands/rpop/

        The pipeline value is None if the key doesn't exist. If a count
        is set a list with values in pop order is set as pipeline value (even
        if it is set to one). If count is not set (default or None) the single
        value that got popped is set as pipeline value.

        Args:
            key (str): The key.

            count (int | None, optional): The number values to pop.
            Defaults to a single value.
        """
        raise NotImplementedError()

    def lrange(self, key: str, start: int, stop: int) -> None:
        """
        Returns a number of values from the list specified by the given range.
        Negative numbers are interpreted as index from the back of the list.
        Out of range indices are ignored, potentially returning an empty list.

        See also the redis documentation: https://redis.io/commands/lrange/

        The pipeline value is the resulting elements.

        Args:
            key (str): The key.

            start (int): The start index.

            stop (int): The stop index (inclusive).
        """
        raise NotImplementedError()

    def lset(self, key: str, index: int, value: str) -> None:
        """
        Sets the value of a list element at the given index. The index can be
        negative to index from the back. The function will raise if the index
        is out of bounds.

        See also the redis documentation: https://redis.io/commands/lset/

        The pipeline value is None.

        Args:
            key (str): The key.

            index (int): The index. Can be negative to index from the back.

            value (str): The value.
        """
        raise NotImplementedError()

    def lindex(self, key: str, index: int) -> None:
        """
        Retrieves the value of a list element at the given index. The index can
        be negative to index from the back. None is returned for an out of
        bounds index.

        See also the redis documentation: https://redis.io/commands/lindex/

        The pipeline value is the value at the index or None if the index was
        out of bounds.

        Args:
            key (str): The key.

            index (int): The index. Can be negative to index from the back.
        """
        raise NotImplementedError()

    def llen(self, key: str) -> None:
        """
        Computes the length of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/llen/

        The length of the list is set as pipeline value.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def zadd(self, key: str, mapping: dict[str, float]) -> None:
        """
        Adds elements to the sorted set associated with the key.

        See also the redis documentation: https://redis.io/commands/zadd/

        NOTE: not all setting modes are implemented yet.

        The number of new members is set as pipeline value.

        Args:
            key (str): The key.
            mapping (dict[str, float]): A dictionary with values and scores.
        """
        raise NotImplementedError()

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> None:
        """
        Pops a number of members of the sorted set associated with the given
        key with the highest scores.

        See also the redis documentation: https://redis.io/commands/zpopmax/

        The members with their associated scores in pop order is set as
        pipeline value.

        Args:
            key (str): The key.

            count (int, optional): The number of members to remove.
            Defaults to 1.
        """
        raise NotImplementedError()

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> None:
        """
        Pops a number of members of the sorted set associated with the given
        key with the lowest scores.

        See also the redis documentation: https://redis.io/commands/zpopmin/

        The members with their associated scores in pop order is set as
        pipeline value.

        Args:
            key (str): The key.

            count (int, optional): The number of members to remove.
            Defaults to 1.
        """
        raise NotImplementedError()

    def zrange(self, key: str, start: int, stop: int) -> None:
        """
        Returns a number of values from the sorted set specified by the given
        range. As of now the indices are based on the order of the set.
        Negative numbers are interpreted as index from the back of the set.
        Out of range indices are ignored, potentially returning an empty set.

        See also the redis documentation: https://redis.io/commands/zrange/

        NOTE: not all modes are implemented yet.

        The members names are set as pipeline value.

        Args:
            key (str): The key.

            start (int): The start index.

            stop (int): The stop index (inclusive).
        """
        raise NotImplementedError()

    def zcard(self, key: str) -> None:
        """
        Computes the cardinality of the sorted set associated with the given
        key.

        See also the redis documentation: https://redis.io/commands/zcard/

        The number of members in the set is set as pipeline value.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def hset(self, key: str, mapping: dict[str, str]) -> None:
        """
        Sets a mapping for the given hash.

        See also the redis documentation: https://redis.io/commands/hset/

        The pipeline value is set to the number of fields added.

        Args:
            key (str): The key.

            mapping (dict[str, str]): The field value pairs to be set.
        """
        raise NotImplementedError()

    def hdel(self, key: str, *fields: str) -> None:
        """
        Deletes fields from the given hash.

        See also the redis documentation: https://redis.io/commands/hdel/

        The pipeline value is set to the number of fields that got deleted
        (excluding fields that did not exist).

        Args:
            key (str): The key.

            *fields (str): The fields to delete.
        """
        raise NotImplementedError()

    def hget(self, key: str, field: str) -> None:
        """
        Retrieves the value associated with a field of a hash.

        See also the redis documentation: https://redis.io/commands/hget/

        The pipeline value is set to the value of the field or None if the
        field doesn't exist.

        Args:
            key (str): The key.

            field (str): The field.
        """
        raise NotImplementedError()

    def hmget(self, key: str, *fields: str) -> None:
        """
        Retrieves the values associated with given fields of a hash.

        See also the redis documentation: https://redis.io/commands/hmget/

        The pipeline value is set to a dictionary with fields mapping to their
        values. If a field doesn't exist in the hash the value is returned
        as None.

        Args:
            key (str): The key.

            *fields (str): The fields to retrieve.
        """
        raise NotImplementedError()

    def hincrby(self, key: str, field: str, inc: float | int) -> None:
        """
        Interprets a field value of a hash as number and updates the value.

        See also the redis documentation:
        https://redis.io/commands/hincrby/
        https://redis.io/commands/hincrbyfloat/

        The pipeline value is set to the new value of the field.

        Args:
            key (str): The key.

            field (str): The field to interpret as number.

            inc (float | int): The relative numerical change.
        """
        raise NotImplementedError()

    def hkeys(self, key: str) -> None:
        """
        Retrieves the fields of a hash.

        See also the redis documentation: https://redis.io/commands/hkeys/

        The pipeline value is set to a list of all fields of the given hash.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def hvals(self, key: str) -> None:
        """
        Retrieves the values of a hash.

        See also the redis documentation: https://redis.io/commands/hvals/

        The pipeline value is set to a list of all values of the given hash.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def hgetall(self, key: str) -> None:
        """
        Retrieves all fields and values of a hash.

        See also the redis documentation: https://redis.io/commands/hgetall/

        The pipeline value is set to a dictionary with fields mapping to their
        values.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def sadd(self, key: str, *values: str) -> None:
        """
        Add values to the set of the given key.

        See also the redis documentation: https://redis.io/commands/sadd/

        The pipeline value is the number of new items in the set.

        Args:
            key (str): The key.
            *values (str): The values to add.
        """
        raise NotImplementedError()

    def srem(self, key: str, *values: str) -> None:
        """
        Remove values from the set of the given key.

        See also the redis documentation: https://redis.io/commands/srem/

        The pipeline value is the number of existing items removed from the
        set.

        Args:
            key (str): The key.
            *values (str): The values to remove.
        """
        raise NotImplementedError()

    def sismember(self, key: str, value: str) -> None:
        """
        Tests whether the given value is present in the set of the given key.

        See also the redis documentation: https://redis.io/commands/sismember/

        The pipeline value is a boolean. True, if the set contains the value.

        Args:
            key (str): The key.
            value (str): The value.
        """
        raise NotImplementedError()

    def scard(self, key: str) -> None:
        """
        Computes the size of the set given by the key.

        See also the redis documentation: https://redis.io/commands/scard/

        The pipeline value is the number of elements in the set.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def smembers(self, key: str) -> None:
        """
        Returns all elements of the set given by the key.

        See also the redis documentation: https://redis.io/commands/smembers/

        The pipeline value is a set of all elements of the set.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()


class RedisAPI:
    """The redis API."""
    def exists(self, *keys: str) -> int:
        """
        Determines whether specified keys exist.

        See also the redis documentation: https://redis.io/commands/exists/

        Args:
            *keys (str): The keys.

        Returns:
            int: The number of keys that exist.
        """
        raise NotImplementedError()

    def delete(self, *keys: str) -> int:
        """
        Deletes keys.

        See also the redis documentation: https://redis.io/commands/del/

        Args:
            *keys (str): The keys.

        Returns:
            int: The number of keys that got removed.
        """
        raise NotImplementedError()

    def key_type(self, key: str) -> KeyType | None:
        """
        The type of the given key if it exists.

        See also the redis documentation: https://redis.io/commands/type/

        Args:
            key (str): The key.

        Returns:
            KeyType | None: The key type or None if the key does not exist.
        """
        raise NotImplementedError()

    def scan(
            self,
            cursor: int,
            *,
            match: str | None = None,
            count: int | None = None,
            filter_type: KeyType | None = None) -> tuple[int, list[str]]:
        """
        Scans the keys of the database.

        See also the redis documentation: https://redis.io/commands/scan/

        Args:
            cursor (int): The cursor. This value is either 0 (for starting a
                scan) or a value returned previously by this function.
            match (str | None, optional): Filters the keys according to a redis
                match string. Defaults to None.
            count (int | None, optional): Estimate of expected returned keys
                in one call. The actual number returned might be different.
                Defaults to None.
            filter_type (KeyType | None, optional): Filters by the key type.
                Defaults to None.

        Returns:
            tuple[int, list[str]]: A tuple of the new cursor and the current
                keys. If the new cursor is 0 the iteration ends.
        """
        raise NotImplementedError()

    def iter_keys(
            self,
            *,
            match: str | None = None,
            filter_type: KeyType | None = None) -> Iterable[str]:
        """
        Iterates matching keys. This is a more streamlined interface to scan.

        See also the redis documentation: https://redis.io/commands/scan/

        Args:
            match (str | None, optional): Filters the keys according to a redis
                match string. Defaults to None.
            filter_type (KeyType | None, optional): Filters by the key type.
                Defaults to None.

        Yields:
            str: The keys of this query. Duplicate keys might get returned.
        """
        cursor = 0
        count = 10
        while True:
            cursor, keys = self.scan(
                cursor,
                match=match,
                count=count,
                filter_type=filter_type)
            yield from keys
            if cursor == 0:
                break
            count = int(min(1000, count * 2))

    def keys_block(
            self,
            *,
            match: str | None = None,
            filter_type: KeyType | None = None) -> list[str]:
        """
        Retrieves all matching keys. This method blocks the full database until
        all keys are returned. In most cases it is better to use the
        non-blocking function instead.

        See also the redis documentation: https://redis.io/commands/keys/

        Args:
            match (str | None, optional): Filters the keys according to a redis
                match string. Defaults to None.
            filter_type (KeyType | None, optional): Filters by the key type.
                Defaults to None.

        Returns:
            list[str]: The list of unique matching keys.
        """
        raise NotImplementedError()

    def keys(
            self,
            *,
            match: str | None = None,
            filter_type: KeyType | None = None,
            block: bool = True) -> set[str]:
        """
        Retrieves all matching keys.

        See also the redis documentation: https://redis.io/commands/scan/ and
            https://redis.io/commands/keys/

        Args:
            match (str | None, optional): Filters the keys according to a redis
                match string. Defaults to None.
            filter_type (KeyType | None, optional): Filters by the key type.
                Defaults to None.
            block (bool, optional): Whether to block the full database while
                retrieving the matching keys. Defaults to True.

        Returns:
            set[str]: The set of unique matching keys.
        """
        if block:
            return set(self.keys_block(match=match, filter_type=filter_type))
        return set(self.iter_keys(match=match, filter_type=filter_type))

    def flushall(self) -> None:
        """
        Flushes all keys in the database. Whether the operation is asynchronous
        is up to the implementation.

        See also the redis documentation: https://redis.io/commands/flushall/
        """
        # FIXME: add `block: bool` argument
        raise NotImplementedError()

    @overload
    def set_value(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[True],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> str | None:
        ...

    @overload
    def set_value(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[False],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> bool | None:
        ...

    @overload
    def set_value(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        ...

    def set_value(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        """
        Sets a value for a given key. The value can be scheduled to expire.

        See also the redis documentation: https://redis.io/commands/set/

        Args:
            key (str): The key.

            value (str): The value.

            mode (RSetMode, optional): Under which condition to set the value
                valid values are RSM_ALWAYS, RSM_MISSING, and RSM_EXISTS.
                RSM_MISSING is the equivalent of setting the NX flag.
                RSM_EXISTS is the equivalent of the XX flag. Defaults to
                RSM_ALWAYS.

            return_previous (bool, optional): Whether to return the previous
                value associated with the key. Defaults to False.

            expire_timestamp (datetime.datetime | None, optional): A timestamp
                on when to expire the key. Defaults to None.

            expire_in (float | None, optional): A relative time in seconds on
                when to expire the key. Defaults to None.

            keep_ttl (bool, optional): Whether to keep previous expiration
                times. Defaults to False.

        Returns:
            str | bool | None: The return value depends on the return_previous
                argument.
        """
        raise NotImplementedError()

    def get_value(self, key: str) -> str | None:
        """
        Retrieves the value for the given key.

        See also the redis documentation: https://redis.io/commands/get/

        Args:
            key (str): The key.

        Returns:
            str | None: The value or None if the key does not exists or the
            value has expired.
        """
        raise NotImplementedError()

    def expire(
            self,
            key: str,
            *,
            mode: RExpireMode = REX_ALWAYS,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None) -> bool:
        """
        Sets or removes the expiration time of a key. The expiration can either
        be a timestamp or a relative amount. If no time is set the expiration
        is removed and the key persists.

        See also the redis documentation:
        https://redis.io/commands/expire/
        and https://redis.io/commands/expireat/
        and https://redis.io/commands/expiretime/
        and https://redis.io/commands/persist/
        and https://redis.io/commands/pexpire/
        and https://redis.io/commands/pexpireat/
        and https://redis.io/commands/pexpiretime/

        Args:
            key (str): The key.

            mode (RExpireMode, optional): The expiration mode. Defaults to
                REX_ALWAYS. The mode is ignored if all expirations are None.

            expire_timestamp (datetime.datetime | None, optional): A timestamp
                on when to expire the key. Defaults to None.

            expire_in (float | None, optional): A relative time in seconds on
                when to expire the key. Defaults to None.

        Returns:
            bool: Whether a change in expiration occurred.
        """
        raise NotImplementedError()

    def ttl(self, key: str) -> float | None:
        """
        Retrieves the time-to-live for a key.

        See also the redis documentation:
        https://redis.io/commands/ttl/ and https://redis.io/commands/pttl/

        Args:
            key (str): The key.

        Returns:
            float | None: The time to live in seconds with millisecond
                precision. If the key is persistent the time is `<= 0.0`.
                If the key does not exist None is returned.
        """
        raise NotImplementedError()

    def incrby(self, key: str, inc: float | int) -> float:
        """
        Updates the value associated with the given key by a relative amount.
        The value is interpreted as number. If the value doesn't exist zero is
        used as starting point.

        See also the redis documentation:
        https://redis.io/commands/incrby/
        https://redis.io/commands/incrbyfloat/

        Args:
            key (str): The key.

            inc (float | int): The relative change.

        Raises:
            ValueError: If the value cannot be interpreted as float.

        Returns:
            float: The new value as float.
        """
        raise NotImplementedError()

    def lpush(self, key: str, *values: str) -> int:
        """
        Pushes values to the left side of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/lpush/

        Args:
            key (str): The key.

            *values (str): The values to push.

        Returns:
            int: The length of the list after the push.
        """
        raise NotImplementedError()

    def rpush(self, key: str, *values: str) -> int:
        """
        Pushes values to the right side of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/rpush/

        Args:
            key (str): The key.

            *values (str): The values to push.

        Returns:
            int: The length of the list after the push.
        """
        raise NotImplementedError()

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        """
        Pops a number of values from the left side of the list associated with
        the key.

        See also the redis documentation: https://redis.io/commands/lpop/

        Args:
            key (str): The key.

            count (int | None, optional): The number values to pop.
            Defaults to a single value.

        Returns:
            str | list[str] | None: None if the key doesn't exist. If a count
            is set a list with values in pop order is returned (even if it is
            set to one). If count is not set (default or None) the single value
            that got popped is returned.
        """
        raise NotImplementedError()

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        """
        Pops a number of values from the right side of the list associated with
        the key.

        See also the redis documentation: https://redis.io/commands/rpop/

        Args:
            key (str): The key.

            count (int | None, optional): The number values to pop.
            Defaults to a single value.

        Returns:
            str | list[str] | None: None if the key doesn't exist. If a count
            is set a list with values in pop order is returned (even if it is
            set to one). If count is not set (default or None) the single value
            that got popped is returned.
        """
        raise NotImplementedError()

    def lrange(self, key: str, start: int, stop: int) -> list[str]:
        """
        Returns a number of values from the list specified by the given range.
        Negative numbers are interpreted as index from the back of the list.
        Out of range indices are ignored, potentially returning an empty list.

        See also the redis documentation: https://redis.io/commands/lrange/

        Args:
            key (str): The key.

            start (int): The start index.

            stop (int): The stop index (inclusive).

        Returns:
            list[str]: The elements.
        """
        raise NotImplementedError()

    def lset(self, key: str, index: int, value: str) -> None:
        """
        Sets the value of a list element at the given index. The index can be
        negative to index from the back. The function will raise if the index
        is out of bounds.

        See also the redis documentation: https://redis.io/commands/lset/

        Args:
            key (str): The key.

            index (int): The index. Can be negative to index from the back.

            value (str): The value.
        """
        raise NotImplementedError()

    def lindex(self, key: str, index: int) -> str | None:
        """
        Retrieves the value of a list element at the given index. The index can
        be negative to index from the back. None is returned for an out of
        bounds index.

        See also the redis documentation: https://redis.io/commands/lindex/

        Args:
            key (str): The key.

            index (int): The index. Can be negative to index from the back.

        Returns:
            str | None: The value at the index or None if the index was out of
                bounds.
        """
        raise NotImplementedError()

    def llen(self, key: str) -> int:
        """
        Computes the length of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/llen/

        Args:
            key (str): The key.

        Returns:
            int: The length of the list.
        """
        raise NotImplementedError()

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        """
        Adds elements to the sorted set associated with the key.

        See also the redis documentation: https://redis.io/commands/zadd/

        NOTE: not all setting modes are implemented yet.

        Args:
            key (str): The key.
            mapping (dict[str, float]): A dictionary with values and scores.

        Returns:
            int: The number of new members.
        """
        raise NotImplementedError()

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        """
        Pops a number of members of the sorted set associated with the given
        key with the highest scores.

        See also the redis documentation: https://redis.io/commands/zpopmax/

        Args:
            key (str): The key.

            count (int, optional): The number of members to remove.
            Defaults to 1.

        Returns:
            list[tuple[str, float]]: The members with their associated scores
            in pop order.
        """
        raise NotImplementedError()

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        """
        Pops a number of members of the sorted set associated with the given
        key with the lowest scores.

        See also the redis documentation: https://redis.io/commands/zpopmin/

        Args:
            key (str): The key.

            count (int, optional): The number of members to remove.
            Defaults to 1.

        Returns:
            list[tuple[str, float]]: The members with their associated scores
            in pop order.
        """
        raise NotImplementedError()

    def zrange(self, key: str, start: int, stop: int) -> list[str]:
        """
        Returns a number of values from the sorted set specified by the given
        range. As of now the indices are based on the order of the set.
        Negative numbers are interpreted as index from the back of the set.
        Out of range indices are ignored, potentially returning an empty set.

        See also the redis documentation: https://redis.io/commands/zrange/

        NOTE: not all modes are implemented yet.

        Args:
            key (str): The key.

            start (int): The start index.

            stop (int): The stop index (inclusive).

        Returns:
            list[str]: The members names.
        """
        raise NotImplementedError()

    def zcard(self, key: str) -> int:
        """
        Computes the cardinality of the sorted set associated with the given
        key.

        See also the redis documentation: https://redis.io/commands/zcard/

        Args:
            key (str): The key.

        Returns:
            int: The number of members in the set.
        """
        raise NotImplementedError()

    def hset(self, key: str, mapping: dict[str, str]) -> int:
        """
        Sets a mapping for the given hash.

        See also the redis documentation: https://redis.io/commands/hset/

        Args:
            key (str): The key.

            mapping (dict[str, str]): The field value pairs to be set.

        Returns:
            int: The number of fields added.
        """
        raise NotImplementedError()

    def hdel(self, key: str, *fields: str) -> int:
        """
        Deletes fields from the given hash.

        See also the redis documentation: https://redis.io/commands/hdel/

        Args:
            key (str): The key.

            *fields (str): The fields to delete.

        Returns:
            int: The number of fields that got deleted (excluding fields that
            did not exist).
        """
        raise NotImplementedError()

    def hget(self, key: str, field: str) -> str | None:
        """
        Retrieves the value associated with a field of a hash.

        See also the redis documentation: https://redis.io/commands/hget/

        Args:
            key (str): The key.

            field (str): The field.

        Returns:
            str | None: The value of the field or None if the field doesn't
            exist.
        """
        raise NotImplementedError()

    def hmget(self, key: str, *fields: str) -> dict[str, str | None]:
        """
        Retrieves the values associated with given fields of a hash.

        See also the redis documentation: https://redis.io/commands/hmget/

        Args:
            key (str): The key.

            *fields (str): The fields to retrieve.

        Returns:
            dict[str, str | None]: A dictionary with fields mapping to their
            values. If a field doesn't exist in the hash the value is returned
            as None.
        """
        raise NotImplementedError()

    def hincrby(self, key: str, field: str, inc: float | int) -> float:
        """
        Interprets a field value of a hash as number and updates the value.

        See also the redis documentation:
        https://redis.io/commands/hincrby/
        https://redis.io/commands/hincrbyfloat/

        Args:
            key (str): The key.

            field (str): The field to interpret as number.

            inc (float | int): The relative numerical change.

        Returns:
            float: The new value of the field.
        """
        raise NotImplementedError()

    def hkeys(self, key: str) -> list[str]:
        """
        Retrieves the fields of a hash.

        See also the redis documentation: https://redis.io/commands/hkeys/

        Args:
            key (str): The key.

        Returns:
            list[str]: All fields of the given hash.
        """
        raise NotImplementedError()

    def hvals(self, key: str) -> list[str]:
        """
        Retrieves the values of a hash.

        See also the redis documentation: https://redis.io/commands/hvals/

        Args:
            key (str): The key.

        Returns:
            list[str]: All values of the given hash.
        """
        raise NotImplementedError()

    def hgetall(self, key: str) -> dict[str, str]:
        """
        Retrieves all fields and values of a hash.

        See also the redis documentation: https://redis.io/commands/hgetall/

        Args:
            key (str): The key.

        Returns:
            dict[str, str]: A dictionary with fields mapping to their values.
        """
        raise NotImplementedError()

    def sadd(self, key: str, *values: str) -> int:
        """
        Add values to the set of the given key.

        See also the redis documentation: https://redis.io/commands/sadd/

        Args:
            key (str): The key.
            *values (str): The values to add.

        Returns:
            int: The number of new items in the set.
        """
        raise NotImplementedError()

    def srem(self, key: str, *values: str) -> int:
        """
        Remove values from the set of the given key.

        See also the redis documentation: https://redis.io/commands/srem/

        Args:
            key (str): The key.
            *values (str): The values to remove.

        Returns:
            int: The number of existing items removed from the set.
        """
        raise NotImplementedError()

    def sismember(self, key: str, value: str) -> bool:
        """
        Tests whether the given value is present in the set of the given key.

        See also the redis documentation: https://redis.io/commands/sismember/

        Args:
            key (str): The key.
            value (str): The value.

        Returns:
            bool: True, if the set contains the value.
        """
        raise NotImplementedError()

    def scard(self, key: str) -> int:
        """
        Computes the size of the set given by the key.

        See also the redis documentation: https://redis.io/commands/scard/

        Args:
            key (str): The key.

        Returns:
            int: The number of elements in the set.
        """
        raise NotImplementedError()

    def smembers(self, key: str) -> set[str]:
        """
        Returns all elements of the set given by the key.

        See also the redis documentation: https://redis.io/commands/smembers/

        Args:
            key (str): The key.

        Returns:
            set[str]: All elements of the set.
        """
        raise NotImplementedError()

    def publish(self, key: str, msg: str) -> None:
        """
        Publishes a message on a pubsub channel.

        See also the redis documentation: https://redis.io/commands/publish/

        Args:
            key (str): The pubsub key.
            msg (str): The message.
        """
        raise NotImplementedError()

    def wait_for(
            self,
            key: str,
            predicate: Callable[[], T],
            timeout: float | None) -> T | None:
        """
        Waits on a pubsub channel until a certain condition is met. This
        ignores messages sent on the channel but instead checks the provided
        condition once a message has been received. If the condition is
        satisfied, its result is returned.

        Args:
            key (str): The pubsub key.
            predicate (Callable[[], T]): If the result of this condition can
                be converted to `True` via `bool` the condition is considered
                satisfied and its result is returned.
            timeout (float | None): If the predicate has not been fullfilled
                within the time set by the timeout in seconds the function
                returns None. If timeout is set to None the function will wait
                indefinitely.

        Returns:
            T | None: The result of the condition is returned or None if the
                function timed out.
        """
        raise NotImplementedError()


class RedisClientAPI(RedisAPI):
    """This class enriches the redis API with pipeline and script
    functionality."""
    @contextlib.contextmanager
    def pipeline(self) -> Iterator[PipelineAPI]:
        """
        Starts a redis pipeline. When leaving the resource block the pipeline
        is executed automatically and the results are discarded. If you need
        the results call execute on the pipeline object.

        Yields:
            PipelineAPI: The pipeline.
        """
        raise NotImplementedError()

    def register_script(self, ctx: FnContext) -> ExecFunction:
        """
        Registers a script that can be executed in this redis runtime.

        Args:
            ctx (FnContext): The script to register.

        Returns:
            ExecFunction: A python that can be called to execute the script.
        """
        raise NotImplementedError()



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A hash based cache. Consumers will be notified as soon as a value is
available if any computation is already ongoing. The cache should be used on
a memory limited redis backend."""
from collections.abc import Callable
from typing import Generic, TypeVar

from redipy.api import RedisClientAPI, RSM_MISSING


K = TypeVar('K')
V = TypeVar('V')


class RCache(Generic[K, V]):  # pylint: disable=too-few-public-methods
    """A redis based cache with in-flight computation detection."""
    def __init__(
            self,
            rt: RedisClientAPI,
            *,
            prefix: str,
            hasher: Callable[[K], str],
            compute: Callable[[K], V],
            value_store: Callable[[V], str],
            value_read: Callable[[str], V]) -> None:
        """
        Creates a redis based cache.

        Args:
            rt (RedisClientAPI): The redis client.

            prefix (str): The key prefix.

            hasher (Callable[[K], str]): Function to create a hash for a given
                key.

            compute (Callable[[K], V]): Compute the actual value for a given
                key.

            value_store (Callable[[V], str]): Convert the value into a string
                for storing. The string must never be empty.

            value_read (Callable[[str], V]): Convert a string into a value for
                retrieving.
        """
        self._rt = rt
        self._prefix = prefix
        self._hasher = hasher
        self._compute = compute
        self._value_store = value_store
        self._value_read = value_read

    def _redis_key(self, hash_str: str) -> str:
        prefix = self._prefix
        if prefix:
            prefix = f"{prefix}:"
        return f"{prefix}{hash_str}"

    def get_value(self, key: K, *, timeout: float = 300.0) -> V:
        """
        Retrieves the value of the given key. If the value is not already
        cached it is computed. If a value is being computed elsewhere at the
        moment the function call blocks until the computation is complete.
        During this, if timeout is reached, the computation will start again.

        Args:
            key (K): The key.

            timeout (float, optional): Timeout before starting to compute the
                value ourself if another computation was already in-flight
                in seconds. Defaults to 300.0.

        Raises:
            ValueError: If the value string is the empty string after
                converting the result.

        Returns:
            V: The result value.
        """
        rt = self._rt
        hash_str = self._hasher(key)
        rkey = self._redis_key(hash_str)
        value = rt.get_value(rkey)
        if value is not None:
            if value:
                return self._value_read(value)
            # someone else is computing the value
            value = rt.wait_for(rkey, lambda: rt.get_value(rkey), timeout)
            if value:
                return self._value_read(value)
        # we have to compute the value
        rt.set_value(rkey, "", mode=RSM_MISSING)
        res = self._compute(key)
        res_str = self._value_store(res)
        if not res_str:
            raise ValueError(f"compute for {key=} with {res=} mapping to ''")
        rt.set_value(rkey, res_str)
        rt.publish(rkey, hash_str)
        return res



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines all execution graph node types of statements. A statement can be
executed alone and usually has side-effects."""
from typing import Literal, TYPE_CHECKING, TypedDict

from redipy.graph.expr import ExprObj, IndexObj, RefIdObj, VarObj


if TYPE_CHECKING:
    from redipy.graph.seq import SequenceObj


AssignmentObj = TypedDict('AssignmentObj', {
    "kind": Literal["assign", "declare"],
    "assign": RefIdObj,
    "value": ExprObj,
})
"""Assigns an expression to a reference."""
AssignAtObj = TypedDict('AssignAtObj', {
    "kind": Literal["assign_at"],
    "assign": RefIdObj,
    "index": ExprObj,
    "value": ExprObj,
})
"""Assigns an expression to an index of an array reference."""
AssignKeyObj = TypedDict('AssignKeyObj', {
    "kind": Literal["assign_key"],
    "assign": RefIdObj,
    "key": ExprObj,
    "value": ExprObj,
})
"""Assigns an expression to a key of a dictionary reference."""
StmtObj = TypedDict('StmtObj', {
    "kind": Literal["stmt"],
    "expr": ExprObj,
})
"""Executes an expression as statement."""
BranchObj = TypedDict('BranchObj', {
    "kind": Literal["branch"],
    "condition": ExprObj,
    "then": 'SequenceObj',
    "else": 'SequenceObj',
})
"""A branch with subsequences whether an expression was truthy or falsey."""
ForLoopObj = TypedDict('ForLoopObj', {
    "kind": Literal["for"],
    "array": ExprObj,
    "index": IndexObj,
    "value": VarObj,
    "body": 'SequenceObj',
})
"""A loop iterating through an array."""
WhileObj = TypedDict('WhileObj', {
    "kind": Literal["while"],
    "condition": ExprObj,
    "body": 'SequenceObj',
})
"""A loop executing until the condition fails."""
ReturnObj = TypedDict('ReturnObj', {
    "kind": Literal["return"],
    "value": ExprObj | None,
})
"""Sets the return value of the current stack frame."""
CommandObj = (
    AssignmentObj
    | AssignAtObj
    | AssignKeyObj
    | StmtObj
    | BranchObj
    | ForLoopObj
    | WhileObj
    | ReturnObj
)
"""Defines all statements."""



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lua helper functions for patching lua function calls."""
from redipy.plugin import HelperFunction


class HPairlistScoresFn(HelperFunction):
    """Converts an alternating list of keys and values into a list of pairs."""
    @staticmethod
    def name() -> str:
        return "pairlist_scores"

    @staticmethod
    def args() -> str:
        return "arr"

    @staticmethod
    def body() -> str:
        return r"""
            local res = {}
            local key = nil
            for ix, value in ipairs(arr) do
                if ix % 2 == 1 then
                    key = value
                else
                    res[#res + 1] = {key, tonumber(value)}
                end
            end
            return res
        """


class HPairlistDictFn(HelperFunction):
    """Converts an alternating list of keys and values into a dictionary."""
    @staticmethod
    def name() -> str:
        return "pairlist_dict"

    @staticmethod
    def args() -> str:
        return "arr"

    @staticmethod
    def body() -> str:
        return r"""
            local res = {}
            local key = nil
            for _, value in ipairs(arr) do
                if key ~= nil then
                    res[key] = value
                    key = nil
                else
                    key = value
                end
            end
            return res
        """


class HKeyValDictFn(HelperFunction):
    """Constructs a dictionary from values and variadic keys."""
    @staticmethod
    def name() -> str:
        return "keyval_dict"

    @staticmethod
    def args() -> str:
        return "values, ..."

    @staticmethod
    def body() -> str:
        return r"""
            local res = {}
            for ix, key in ipairs(arg) do
                res[key] = values[ix] or cjson.null
            end
            return res
        """


class HNilOrIndexFn(HelperFunction):
    """Updates an index from 1-based (lua) to 0-based (python) if the value is
    not None (nil)."""
    @staticmethod
    def name() -> str:
        return "nil_or_index"

    @staticmethod
    def args() -> str:
        return "val"

    @staticmethod
    def body() -> str:
        return r"""
            if val ~= nil then
                val = val - 1
            end
            return val
        """


class HAsIntStrFn(HelperFunction):
    """Returns the integer value of a number."""
    @staticmethod
    def name() -> str:
        return "asintstr"

    @staticmethod
    def args() -> str:
        return "val"

    @staticmethod
    def body() -> str:
        return r"""
            return math.floor(val)
        """



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test keys operations."""


import re
from collections.abc import Callable, Iterable
from test.util import get_setup
from typing import Any, cast

import pytest

from redipy.api import KeyType, PipelineAPI
from redipy.backend.runtime import Runtime
from redipy.main import Redis
from redipy.util import convert_pattern


KEY_TYPE_REG: list[KeyType] = [
    "string",
    "list",
    "set",
    "zset",
    "hash",
]


KEY_TYPE_MODES: list[KeyType | None] = [
    "string",
    "list",
    "set",
    "zset",
    "hash",
    None,
]


DEFAULTS: dict[
        KeyType, Callable[[Runtime | PipelineAPI, str, int], object]] = {
    "string": lambda rt, key, ix: rt.set_value(key, f"v{ix}"),
    "list": lambda rt, key, ix: rt.rpush(key, f"v{ix}"),
    "set": lambda rt, key, ix: rt.sadd(key, f"v{ix}"),
    "zset": lambda rt, key, ix: rt.zadd(key, {f"v{ix}": ix}),
    "hash": lambda rt, key, ix: rt.hset(key, {"value": f"v{ix}"}),
}


CHECKS: dict[KeyType, Callable[[Runtime, str, int], bool]] = {
    "string": lambda rt, key, ix: rt.get_value(key) == f"v{ix}",
    "list": lambda rt, key, ix: set(rt.lrange(key, 0, -1)) == {f"v{ix}"},
    "set": lambda rt, key, ix: rt.smembers(key) == {f"v{ix}"},
    "zset": lambda rt, key, ix: rt.zrange(key, 0, -1) == [f"v{ix}"],
    "hash": lambda rt, key, ix: rt.hgetall(key) == {"value": f"v{ix}"},
}


PIPE_CHECKS: dict[KeyType, Callable[[PipelineAPI, str], None]] = {
    "string": lambda pipe, key: pipe.get_value(key),
    "list": lambda pipe, key: pipe.lrange(key, 0, -1),
    "set": lambda pipe, key: pipe.smembers(key),
    "zset": lambda pipe, key: pipe.zrange(key, 0, -1),
    "hash": lambda pipe, key: pipe.hgetall(key),
}


PIPE_EXPECTED: dict[KeyType, Callable[[Any, int], bool]] = {
    "string": lambda res, ix: res == f"v{ix}",
    "list": lambda res, ix: set(res) == {f"v{ix}"},
    "set": lambda res, ix: res == {f"v{ix}"},
    "zset": lambda res, ix: res == [f"v{ix}"],
    "hash": lambda res, ix: res == {"value": f"v{ix}"},
}


@pytest.mark.parametrize("types", KEY_TYPE_MODES)
@pytest.mark.parametrize("k_add", [None, 11])
@pytest.mark.parametrize("k_del", [None, 7])
@pytest.mark.parametrize("match", [None, "k1*", "k???"])
@pytest.mark.parametrize("count", [30, 200, 500, 1000, 2000, 10000, 100000])
@pytest.mark.parametrize("rt_lua", [False, True])
def test_scan(
        types: KeyType | None,
        k_add: int | None,
        k_del: int | None,
        match: str | None,
        count: int,
        rt_lua: bool) -> None:
    """
    Test scan command with concurrent writes and deletes.

    Args:
        types (KeyType | None): Which key types to use.
        k_add (int | None): Which keys to add late.
        k_del (int | None): Which keys to delete.
        match (str | None): The filter pattern.
        count (int): How many keys to generate.
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    rt = get_setup("test_keys", rt_lua)

    if count > 200 and (k_add is None or k_del is None):
        # NOTE: those tests wouldn't add anything interesting
        return
    if count > 2000 and (types is not None or match is None):
        # NOTE: we reduce test cases for large data
        return
    if rt_lua and count >= 10000:
        # NOTE: redis backend tests get really slow with many keys
        return

    all_keys: dict[str, KeyType] = {}
    keys: dict[str, KeyType] = {}
    maybe: dict[str, KeyType] = {}
    later: list[tuple[KeyType, str]] = []

    def gen(
            start: int,
            stop: int,
            step: int = 1) -> Iterable[tuple[KeyType, str]]:
        yield from ((
            KEY_TYPE_REG[ix % len(KEY_TYPE_REG)] if types is None else types,
            f"k{ix}",
        ) for ix in range(start, stop, step))

    def extract(key: str) -> int:
        return int(key[1:])

    for key_type, key in gen(0, count):
        ix = extract(key)
        if k_add is not None and ix % k_add == 0:
            later.append((key_type, key))
            continue
        DEFAULTS[key_type](rt, key, ix)
        keys[key] = key_type
        all_keys[key] = key_type
    assert len(keys) + len(later) == count
    total: set[str] = set()

    def cur_max() -> int:
        if not total:
            return -1
        return max((extract(key) for key in total))

    def cond_op(
            arr: Iterable[tuple[KeyType, str]],
            cond: Callable[[int], bool],
            *,
            is_add: bool) -> None:
        for cur_type, cur_key in arr:
            cur_ix = extract(cur_key)
            if not cond(cur_ix):
                continue
            maybe[cur_key] = cur_type
            if is_add:
                DEFAULTS[cur_type](rt, cur_key, cur_ix)
                # NOTE: we do not add to our reference since scan doesn't
                # guarantee those keys
                all_keys[cur_key] = cur_type
            else:
                exp_type = keys.pop(cur_key, None)
                if exp_type is None:
                    # NOTE: key might have never been added
                    rt.delete(cur_key)
                    assert rt.exists(cur_key) == 0
                    assert rt.key_type(cur_key) is None
                else:
                    assert rt.exists(cur_key) == 1
                    assert rt.key_type(cur_key) == exp_type
                    assert rt.delete(cur_key) == 1
    iters = 0
    cursor = 0
    scan_count = 50
    while True:
        cursor, partial = rt.scan(cursor, match=match, count=scan_count)
        total.update(partial)
        if cursor == 0:
            break
        iters += 1
        if iters == 3:
            cmax_add = cur_max()
            cond_op(later, lambda ix: ix < cmax_add, is_add=True)
        elif iters == 5 and k_del is not None:
            cmax_del = cur_max()
            cond_op((gen(0, cmax_del, k_del)), lambda _: True, is_add=False)
        elif iters == 7:
            cond_op(later, lambda _: True, is_add=True)
        elif iters == 11 and k_del is not None:
            cond_op(gen(0, count, k_del), lambda _: True, is_add=False)
        elif iters == 12:
            scan_count = 1000  # NOTE: return the rest a bit quicker

    pat: re.Pattern | None = None
    if match:
        _, pat = convert_pattern(match)
    for ref_key, ref_type in keys.items():
        if pat is not None and not pat.match(ref_key):
            assert ref_key not in total
            continue
        assert ref_key in total
        assert rt.exists(ref_key) > 0
        assert rt.key_type(ref_key) == ref_type
        assert CHECKS[ref_type](rt, ref_key, extract(ref_key))
        assert rt.delete(ref_key) == 1
    for m_key, m_type in maybe.items():
        if pat is not None and not pat.match(m_key):
            continue
        if rt.exists(m_key) == 0:
            continue
        assert rt.key_type(m_key) == m_type
        assert CHECKS[m_type](rt, m_key, extract(m_key))
        assert rt.delete(m_key) == 1
    for rem_key in total:
        if rt.exists(rem_key):
            rem_type = rt.key_type(rem_key)
            assert rem_type is not None
            assert CHECKS[rem_type](rt, rem_key, extract(rem_key))
            assert rt.delete(rem_key) == 1
    for final_key in rt.keys(block=True):
        assert final_key in all_keys
        final_type = all_keys[final_key]
        assert rt.exists(final_key) > 0
        assert rt.key_type(final_key) == final_type
        assert CHECKS[final_type](rt, final_key, extract(final_key))
        assert rt.delete(final_key) == 1


@pytest.mark.parametrize("types", KEY_TYPE_MODES)
@pytest.mark.parametrize("match", [None, "k1*", "k???"])
@pytest.mark.parametrize("block", [False, True])
@pytest.mark.parametrize("rt_lua", [False, True])
def test_keys(
        types: KeyType | None,
        match: str | None,
        block: bool,
        rt_lua: bool) -> None:
    """
    Test keys command.

    Args:
        types (KeyType | None): Which key types to use.
        match (str | None): The filter pattern.
        block (bool): Whether to use blocking operations.
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    rt = get_setup("test_keys", rt_lua)
    if rt_lua and types is not None:
        return  # FIXME: implement filter_type for redis connections

    rt_alt = get_setup("test_keys_alt", rt_lua)
    # NOTE: this value should never be returned by keys or scan
    rt_alt.set_value("foo", "clobber")

    def gen(
            start: int,
            stop: int,
            step: int = 1) -> Iterable[tuple[KeyType, str]]:
        yield from ((
            KEY_TYPE_REG[ix % len(KEY_TYPE_REG)] if types is None else types,
            f"k{ix}",
        ) for ix in range(start, stop, step))

    def extract(key: str) -> int:
        return int(key[1:])

    keys: dict[str, KeyType] = {}
    assert rt.keys(match=match, block=block, filter_type=types) == set()
    with rt.pipeline() as pipe:
        pipe.scan(0, match=match, filter_type=types)
        pipe.keys(match=match, filter_type=types)
        scan_res, keys_res = pipe.execute()
        # NOTE: scan_res[0] could be != 0 if other keys exist in the db
        assert scan_res[1] == []
        assert keys_res == []
        for key_type, key in gen(0, 100):
            ix = extract(key)
            DEFAULTS[key_type](pipe, key, ix)
            keys[key] = key_type
    assert rt.keys(block=block) == set(keys.keys())

    def for_type(filter_type: KeyType | None) -> None:
        total = rt.keys(block=block, match=match, filter_type=filter_type)
        pat: re.Pattern | None = None
        if match:
            _, pat = convert_pattern(match)

        def get_type_check(ref_type: str) -> Callable[[Any], bool]:
            return lambda res: res == ref_type

        def get_pipe_check(
                ref_key: str, ref_type: KeyType) -> Callable[[Any], bool]:
            expected = PIPE_EXPECTED[ref_type]
            ix = extract(ref_key)
            return lambda res: expected(res, ix)

        checks: list[tuple[Callable[[Any], bool], str]] = []
        with rt.pipeline() as pipe:
            for ref_key, ref_type in keys.items():
                if pat is not None and not pat.match(ref_key):
                    assert ref_key not in total
                    continue
                if filter_type is not None and ref_type != filter_type:
                    assert ref_key not in total
                    continue
                assert ref_key in total
                pipe.exists(ref_key)
                checks.append((lambda res: int(res) > 0, f"exists {ref_key}"))
                pipe.key_type(ref_key)
                checks.append((
                    get_type_check(ref_type),
                    f"type {ref_key} {ref_type}"))
                PIPE_CHECKS[ref_type](pipe, ref_key)
                checks.append((
                    get_pipe_check(ref_key, ref_type),
                    f"check {ref_key} {ref_type}"))
            results = pipe.execute()
        for cur_res, cur_check in zip(results, checks):
            check, check_name = cur_check
            assert check(cur_res), f"{check_name} result: {cur_res}"

    test_types: list[KeyType | None] = [None] if rt_lua else KEY_TYPE_MODES
    for test_type in test_types:
        for_type(test_type)

    with rt.pipeline() as pipe:
        pipe.keys()
        finals_keys: list[str] = cast(list, pipe.execute()[0])

    for final_key in finals_keys:
        assert final_key in keys
        final_type = keys[final_key]
        assert rt.exists(final_key) > 0
        assert rt.key_type(final_key) == final_type
        assert CHECKS[final_type](rt, final_key, extract(final_key))
        assert rt.delete(final_key) == 1

    assert rt.keys(match=match, block=block, filter_type=types) == set()
    with rt.pipeline() as pipe:
        pipe.scan(0, match=match, filter_type=types)
        pipe.keys(match=match, filter_type=types)
        scan_final, keys_final = pipe.execute()
        assert scan_final[1] == []
        assert keys_final == []
    cursor = scan_final[0]
    while cursor != 0:
        cursor, scan_arr = rt.scan(cursor, match=match, filter_type=types)
        assert scan_arr == []

    assert rt_alt.get_value("foo") == "clobber"


@pytest.mark.parametrize("rt_lua", [False, True])
def test_flushall(rt_lua: bool) -> None:
    """
    Test flushall command.

    Args:
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    rt = get_setup("test_flushall", rt_lua)

    rt_alt = Redis(rt=get_setup("test_flushall_alt", rt_lua))
    assert rt_alt.keys() == set()
    # NOTE: this value must not be removed
    rt_alt.set_value("foo", "stayingalive")

    assert rt.keys() == set()
    assert rt_alt.get_value("foo") == "stayingalive"
    rt.flushall()
    assert rt.keys() == set()
    assert rt_alt.get_value("foo") == "stayingalive"

    def gen(
            start: int,
            stop: int,
            step: int = 1) -> Iterable[tuple[KeyType, str]]:
        yield from ((
            KEY_TYPE_REG[ix % len(KEY_TYPE_REG)],
            f"k{ix}",
        ) for ix in range(start, stop, step))

    def extract(key: str) -> int:
        return int(key[1:])

    keys: dict[str, KeyType] = {}
    for key_type, key in gen(0, 100):
        ix = extract(key)
        DEFAULTS[key_type](rt, key, ix)
        keys[key] = key_type

    assert rt.keys() == set(keys.keys())
    assert rt_alt.get_value("foo") == "stayingalive"
    rt.flushall()
    assert rt.keys() == set()
    assert rt_alt.get_value("foo") == "stayingalive"
    assert rt_alt.key_type("foo") == "string"
    assert rt_alt.keys_block() == ["foo"]

    with rt_alt.pipeline() as pipe:
        pipe.delete("foo")
        pipe.key_type("foo")
        pipe.set_value("foo", "bar")
        assert pipe.execute() == [1, None, True]

    assert rt_alt.get_value("foo") == "bar"
    rt_alt.flushall()
    assert rt_alt.keys() == set()
    assert rt_alt.get_value("foo") is None



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines all execution graph node types of sequences. A sequence groups
together statements."""
from typing import Literal, TypedDict

from redipy.graph.cmd import CommandObj


SeqObj = TypedDict('SeqObj', {
    "kind": Literal["seq"],
    "cmds": list[CommandObj],
})
"""A normal sequence that does not add a new stack frame."""
ScriptObj = TypedDict('ScriptObj', {
    "kind": Literal["script"],
    "cmds": list[CommandObj],
    "argv": list[str],
    "keyv": list[str],
})
"""The base sequence of the script."""
FunctionObj = TypedDict('FunctionObj', {
    "kind": Literal["function"],
    "cmds": list[CommandObj],
    "argc": int,
})
"""A sequence for defining custom local functions."""


SequenceObj = SeqObj | ScriptObj | FunctionObj
"""A sequence grouping together statements."""



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests sorted sets."""
from test.util import get_setup, run_code

import pytest

from redipy.backend.backend import ExecFunction
from redipy.main import Redis

# from redipy.symbolic.expr import Strs
# from redipy.symbolic.fun import LogFn, ToJSON
from redipy.symbolic.rset import RedisSet
from redipy.symbolic.seq import FnContext
from redipy.util import lua_fmt


EMPTY_OBJ = r"{}"


LUA_SCRIPT = lua_fmt(f"""
--[[ KEYV
left_set
right_set
]]
--[[ ARGV
swap
]]
local arg_0 = cjson.decode(ARGV[1])  -- swap
local key_0 = (KEYS[1])  -- left_set
local key_1 = (KEYS[2])  -- right_set
local var_0 = cjson.decode("{EMPTY_OBJ}")
for ix_0, val_0 in ipairs(arg_0) do
    var_0[val_0] = true
end
local var_1 = cjson.decode("{EMPTY_OBJ}")
for ix_1, val_1 in ipairs(redis.call("smembers", key_0)) do
    if var_0[val_1] then
        if (not (redis.call("sismember", key_1, val_1) ~= 0)) then
            redis.call("sadd", key_1, val_1)
            redis.call("srem", key_0, val_1)
            var_1[val_1] = true
        end
    end
end
for ix_2, val_2 in ipairs(redis.call("smembers", key_1)) do
    if var_0[val_2] then
        if (not (redis.call("sismember", key_0, val_2) ~= 0)) then
            redis.call("sadd", key_0, val_2)
            redis.call("srem", key_1, val_2)
            var_1[val_2] = true
        end
    end
end
return cjson.encode(var_1)
""")


ASET: set[str] = {
    "a",
    "b",
    "c",
    "d",
}


BSET: set[str] = {
    "b",
    "d",
    "e",
    "f",
}


RUN_TESTS: list[tuple[list[str], set[str] | None]] = [
    # -- a --
    # a
    # b
    # c
    # d
    # -- b --
    # b
    # d
    # e
    # f
    (["a", "b", "c"], {"a", "c"}),
    # -- a --
    # b
    # d
    # -- b --
    # a
    # b
    # c
    # d
    # e
    # f
    (["a", "d", "e", "f"], {"a", "e", "f"}),
    # -- a --
    # a
    # b
    # d
    # e
    # f
    # -- b --
    # b
    # c
    # d
    (["c"], {"c"}),
    # -- a --
    # a
    # b
    # c
    # d
    # e
    # f
    # -- b --
    # b
    # d
    (["d"], None),
]


@pytest.mark.parametrize("rt_lua", [False, True])
def test_rset(rt_lua: bool) -> None:
    """
    Tests sets.

    Args:
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    redis = Redis(rt=get_setup("test_rset", rt_lua, lua_script=LUA_SCRIPT))

    ctx = FnContext()
    swap = ctx.add_arg("swap")

    left_set = RedisSet(ctx.add_key("left_set"))
    right_set = RedisSet(ctx.add_key("right_set"))

    # ctx.add(LogFn("warning", Strs("swap ", ToJSON(swap))))

    lookup = ctx.add_local({})
    rloop, _, rcur = ctx.for_(swap)
    # rloop.add(LogFn("warning", Strs("rcur ", ToJSON(rcur))))
    rloop.add(lookup.set_key(rcur, True))

    res = ctx.add_local({})
    # ctx.add(LogFn("warning", Strs("left_set ", ToJSON(left_set.members()))))
    # ctx.add(
    #     LogFn("warning", Strs("right_set ", ToJSON(right_set.members()))))
    # ctx.add(LogFn("warning", Strs("lookup ", ToJSON(lookup))))
    left_loop, _, cur_left = ctx.for_(left_set.members())
    # left_loop.add(LogFn("warning", Strs("cur_left ", cur_left)))
    # left_loop.add(
    #     LogFn("warning", Strs("get_key ", ToJSON(lookup.get_key(cur_left)))))
    # left_loop.add(
    #     LogFn("warning", Strs("has ", ToJSON(right_set.has(cur_left)))))
    left_then, _ = left_loop.if_(lookup.get_key(cur_left))
    right_empty, _ = left_then.if_(right_set.has(cur_left).not_())
    right_empty.add(right_set.add(cur_left))
    right_empty.add(left_set.remove(cur_left))
    right_empty.add(res.set_key(cur_left, True))
    # right_empty.add(LogFn("warning", Strs("res ", ToJSON(res))))

    right_loop, _, cur_right = ctx.for_(right_set.members())
    # right_loop.add(LogFn("warning", Strs("cur_right ", cur_right)))
    # right_loop.add(LogFn(
    #     "warning",
    #     Strs("get_key ", ToJSON(lookup.get_key(cur_right)))))
    # right_loop.add(
    #     LogFn("warning", Strs("has ", ToJSON(right_set.has(cur_right)))))
    right_then, _ = right_loop.if_(lookup.get_key(cur_right))
    left_empty, _ = right_then.if_(left_set.has(cur_right).not_())
    left_empty.add(left_set.add(cur_right))
    left_empty.add(right_set.remove(cur_right))
    left_empty.add(res.set_key(cur_right, True))
    # left_empty.add(LogFn("warning", Strs("res ", ToJSON(res))))

    ctx.set_return_value(res)

    assert redis.sadd("a", *ASET) == len(ASET)
    assert redis.smembers("a") == ASET
    assert redis.sadd("b", *BSET) == len(BSET)
    assert redis.smembers("b") == BSET
    cset = {"a", "b"}
    assert redis.sadd("c", *cset) == len(cset)
    assert redis.smembers("c") == cset

    def tester(
            runner: ExecFunction,
            vals: list[str]) -> set[str] | None:
        print(f"execute {vals}")
        res = runner(
            keys={"left_set": "a", "right_set": "b"},
            args={"swap": vals})
        if res:
            assert isinstance(res, dict)
            return set(res.keys())
        assert res is None
        return None

    run_code(redis, ctx, tests=RUN_TESTS, tester=tester)
    assert redis.scard("a") == 6
    assert redis.smembers("a") == {"a", "b", "c", "d", "e", "f"}
    assert redis.scard("b") == 2
    assert redis.smembers("b") == {"b", "d"}
    assert redis.sismember("c", "a")
    assert redis.sismember("c", "b")
    assert not redis.sismember("c", "c")
    assert redis.scard("c") == 2

    assert redis.srem("a", "a", "b", "c", "g") == 3
    with redis.pipeline() as pipe:
        pipe.scard("a")
        pipe.smembers("a")
        pipe.srem("a", "d", "e", "f")
        pipe.exists("a")
        pipe.sismember("b", "a")
        pipe.sismember("b", "b")
        pipe.srem("b", "b")
        pipe.scard("b")
        pipe.srem("b", "d")
        pipe.scard("b")
        pipe_res = pipe.execute()
    assert pipe_res[0] == 3  # scard
    assert pipe_res[1] == {"d", "e", "f"}  # smembers
    assert pipe_res[2] == 3  # srem
    assert not pipe_res[3]  # exists
    assert not pipe_res[4]  # sismember
    assert pipe_res[5]  # sismember
    assert pipe_res[6] == 1  # srem
    assert pipe_res[7] == 1  # scard
    assert pipe_res[8] == 1  # srem
    assert pipe_res[9] == 0  # scard



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests various error cases."""

from collections.abc import Callable
from test.util import get_setup
from typing import Any

import pytest

from redipy.api import as_key_type, KeyType, PipelineAPI, RedisAPI
from redipy.main import Redis


KEY_TYPE_REG: list[KeyType] = [
    "string",
    "list",
    "set",
    "zset",
    "hash",
]


DEFAULTS: dict[
        KeyType, Callable[[RedisAPI | PipelineAPI, str, int], object]] = {
    "string": lambda rt, key, ix: rt.set_value(key, f"v{ix}"),
    "list": lambda rt, key, ix: rt.rpush(key, f"v{ix}"),
    "set": lambda rt, key, ix: rt.sadd(key, f"v{ix}"),
    "zset": lambda rt, key, ix: rt.zadd(key, {f"v{ix}": ix}),
    "hash": lambda rt, key, ix: rt.hset(key, {"value": f"v{ix}"}),
}


CHECKS: dict[KeyType, Callable[[RedisAPI, str], Any]] = {
    "string": lambda rt, key: rt.get_value(key),
    "list": lambda rt, key: rt.lrange(key, 0, -1),
    "set": lambda rt, key: rt.smembers(key),
    "zset": lambda rt, key: rt.zrange(key, 0, -1),
    "hash": lambda rt, key: rt.hgetall(key),
}


PIPE_CHECKS: dict[KeyType, Callable[[PipelineAPI, str], None]] = {
    "string": lambda pipe, key: pipe.get_value(key),
    "list": lambda pipe, key: pipe.lrange(key, 0, -1),
    "set": lambda pipe, key: pipe.smembers(key),
    "zset": lambda pipe, key: pipe.zrange(key, 0, -1),
    "hash": lambda pipe, key: pipe.hgetall(key),
}


EXPECTED: dict[KeyType, Callable[[Any, int], bool]] = {
    "string": lambda res, ix: res == f"v{ix}",
    "list": lambda res, ix: set(res) == {f"v{ix}"},
    "set": lambda res, ix: res == {f"v{ix}"},
    "zset": lambda res, ix: res == [f"v{ix}"],
    "hash": lambda res, ix: res == {"value": f"v{ix}"},
}


def test_errors() -> None:
    """Tests various error or edge cases."""
    with pytest.raises(ValueError, match="unknown key type: foo"):
        as_key_type("foo")
    assert as_key_type("none") is None

    with pytest.raises(ValueError, match="unknown backend foo"):
        Redis(backend="foo")  # type: ignore

    with pytest.raises(
            ValueError, match="rt must not be None for custom backend"):
        Redis(backend="custom")

    rt_redis = Redis("custom", rt=get_setup("test_err", rt_lua=True))
    assert rt_redis.maybe_get_redis_runtime() is not None
    assert rt_redis.get_redis_runtime() is not None
    assert rt_redis.maybe_get_memory_runtime() is None
    with pytest.raises(ValueError, match="not a memory runtime"):
        rt_redis.get_memory_runtime()

    rt_mem = Redis("custom", rt=get_setup("test_err", rt_lua=False))
    assert rt_mem.maybe_get_redis_runtime() is None
    with pytest.raises(ValueError, match="not a redis runtime"):
        assert rt_mem.get_redis_runtime() is not None
    assert rt_mem.maybe_get_memory_runtime() is not None
    assert rt_mem.get_memory_runtime() is not None


@pytest.mark.parametrize("rt_lua", [False, True])
def test_rt_errors(rt_lua: bool) -> None:
    """
    Test runtime errors.

    Args:
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    rt = get_setup("test_rt_errors", rt_lua)
    redis = Redis(rt=rt)

    assert redis.scan(0) == (0, [])

    for ix, key_type in enumerate(KEY_TYPE_REG):
        key = f"k{ix}"
        DEFAULTS[key_type](redis, key, ix)

    for key_type_left in KEY_TYPE_REG:
        for ix, key_type_right in enumerate(KEY_TYPE_REG):
            key = f"k{ix}"
            print(
                f"key: {key} "
                f"expected: {key_type_left} actual: {key_type_right}")
            if key_type_left == key_type_right:
                val = CHECKS[key_type_left](redis, key)
                assert EXPECTED[key_type_left](val, ix)
            else:
                with pytest.raises(
                        TypeError,
                        match=r"key.*(ha|i)s a"):
                    res = CHECKS[key_type_left](redis, key)
                    print(
                        f"got: {res} "
                        f"correct: {EXPECTED[key_type_left](res, ix)}")



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A dictionary stack in redis with keys shadowed in stack frames."""
from typing import cast

from redipy.api import RedisClientAPI
from redipy.backend.backend import ExecFunction
from redipy.graph.expr import JSONType
from redipy.symbolic.expr import Strs
from redipy.symbolic.fun import ToIntStr, ToNum
from redipy.symbolic.rhash import RedisHash
from redipy.symbolic.rvar import RedisVar
from redipy.symbolic.seq import FnContext


class RStack:
    """
    A dictionary stack in redis. Keys can be shadowed in stack frames.
    """
    def __init__(self, rt: RedisClientAPI) -> None:
        """
        Creates a dictionary stack for the given redis client.

        Args:
            rt (RedisClientAPI): The redis client.
        """
        self._rt = rt

        self._set_value = self._set_value_script()
        self._get_value = self._get_value_script()
        self._pop_frame = self._pop_frame_script()
        self._get_cascading = self._get_cascading_script()

    def key(self, base: str, name: str) -> str:
        """
        Compute the key.

        Args:
            base (str): The base key.

            name (str): The name.

        Returns:
            str: The key associated with the name.
        """
        return f"{base}:{name}"

    def push_frame(self, base: str) -> None:
        """
        Pushes a new stack frame.

        Args:
            base (str): The base key.
        """
        self._rt.incrby(self.key(base, "size"), 1)

    def pop_frame(self, base: str) -> dict[str, str]:
        """
        Pops the current stack frame and returns its values.

        Args:
            base (str): The base key.

        Returns:
            dict[str, str] | None: The content of the stack frame.
        """
        res = self._pop_frame(
            keys={
                "size": self.key(base, "size"),
                "frame": self.key(base, "frame"),
            },
            args={})
        if res is None:
            return {}
        return cast(dict, res)

    def set_value(self, base: str, field: str, value: str) -> None:
        """
        Set a value in the current stack frame.

        Args:
            base (str): The base key.

            field (str): The field.

            value (str): The value.
        """
        self._set_value(
            keys={
                "size": self.key(base, "size"),
                "frame": self.key(base, "frame"),
            },
            args={"field": field, "value": value})

    def get_value(
            self, base: str, field: str, *, cascade: bool = False) -> JSONType:
        """
        Returns a value from the stack. If the value is not in the current
        stack frame and cascade is set, the value is recursively retrieved
        from the previous stack frames.

        Args:
            base (str): The base key.

            field (str): The field.

            cascade (bool): Whether to recursively inspect all stack frames.

        Returns:
            JSONType: The value.
        """
        if cascade:
            return self._get_cascading(
                keys={
                    "size": self.key(base, "size"),
                    "frame": self.key(base, "frame"),
                },
                args={"field": field})
        return self._get_value(
            keys={
                "size": self.key(base, "size"),
                "frame": self.key(base, "frame"),
            },
            args={"field": field})

    def _set_value_script(self) -> ExecFunction:
        ctx = FnContext()
        rsize = RedisVar(ctx.add_key("size"))
        rframe = RedisHash(Strs(
            ctx.add_key("frame"),
            ":",
            ToIntStr(rsize.get_value(default=0))))
        field = ctx.add_arg("field")
        value = ctx.add_arg("value")
        ctx.add(rframe.hset({
            field: value,
        }))
        ctx.set_return_value(None)
        return self._rt.register_script(ctx)

    def _get_value_script(self) -> ExecFunction:
        ctx = FnContext()
        rsize = RedisVar(ctx.add_key("size"))
        rframe = RedisHash(Strs(
            ctx.add_key("frame"),
            ":",
            ToIntStr(rsize.get_value(default=0))))
        field = ctx.add_arg("field")
        ctx.set_return_value(rframe.hget(field))
        return self._rt.register_script(ctx)

    def _pop_frame_script(self) -> ExecFunction:
        ctx = FnContext()
        rsize = RedisVar(ctx.add_key("size"))
        rframe = RedisHash(Strs(
            ctx.add_key("frame"), ":", ToIntStr(rsize.get_value(default=0))))
        lcl = ctx.add_local(rframe.hgetall())
        ctx.add(rframe.delete())

        b_then, b_else = ctx.if_(ToNum(rsize.get_value(default=0)).gt_(0))
        b_then.add(rsize.incrby(-1))
        b_else.add(rsize.delete())

        ctx.set_return_value(lcl)
        return self._rt.register_script(ctx)

    def _get_cascading_script(self) -> ExecFunction:
        ctx = FnContext()
        rsize = RedisVar(ctx.add_key("size"))
        base = ctx.add_local(ctx.add_key("frame"))
        field = ctx.add_arg("field")
        pos = ctx.add_local(ToNum(rsize.get_value(default=0)))
        res = ctx.add_local(None)
        cur = ctx.add_local(None)
        rframe = RedisHash(cur)

        loop = ctx.while_(res.eq_(None).and_(pos.ge_(0)))
        loop.add(cur.assign(Strs(base, ":", ToIntStr(pos))))
        loop.add(res.assign(rframe.hget(field)))
        loop.add(pos.assign(pos - 1))

        ctx.set_return_value(res)
        return self._rt.register_script(ctx)



# Copyright 2024 Josua Krause
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test expire operations."""


import time
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from datetime import datetime, timedelta
from test.util import get_setup
from typing import Any, TypeAlias

import pytest

from redipy.api import (
    KeyType,
    PipelineAPI,
    REX_ALWAYS,
    REX_EARLIER,
    REX_EXPIRE,
    REX_LATER,
    REX_PERSIST,
    RExpireMode,
)
from redipy.main import Redis
from redipy.util import now


Action: TypeAlias = tuple[Callable[[Any], bool], Callable[[Any], str]]


KEY_TYPE_REG: list[KeyType] = [
    "string",
    "list",
    "set",
    "zset",
    "hash",
]


DEFAULTS: dict[
        KeyType, Callable[[Redis | PipelineAPI, str, int], object]] = {
    "string": lambda rt, key, ix: rt.set_value(key, f"v{ix}"),
    "list": lambda rt, key, ix: rt.rpush(key, f"v{ix}"),
    "set": lambda rt, key, ix: rt.sadd(key, f"v{ix}"),
    "zset": lambda rt, key, ix: rt.zadd(key, {f"v{ix}": ix}),
    "hash": lambda rt, key, ix: rt.hset(key, {"value": f"v{ix}"}),
}


CHECKS: dict[KeyType, Callable[[Redis, str, int], bool]] = {
    "string": lambda rt, key, ix: rt.get_value(key) == f"v{ix}",
    "list": lambda rt, key, ix: set(rt.lrange(key, 0, -1)) == {f"v{ix}"},
    "set": lambda rt, key, ix: rt.smembers(key) == {f"v{ix}"},
    "zset": lambda rt, key, ix: rt.zrange(key, 0, -1) == [f"v{ix}"],
    "hash": lambda rt, key, ix: rt.hgetall(key) == {"value": f"v{ix}"},
}


MISSING: dict[KeyType, Callable[[Redis, str], bool]] = {
    "string": lambda rt, key: rt.get_value(key) is None,
    "list": lambda rt, key: rt.lrange(key, 0, -1) == [],
    "set": lambda rt, key: rt.smembers(key) == set(),
    "zset": lambda rt, key: rt.zrange(key, 0, -1) == [],
    "hash": lambda rt, key: rt.hgetall(key) == {},
}


PIPE_CHECKS: dict[KeyType, Callable[[PipelineAPI, str], None]] = {
    "string": lambda pipe, key: pipe.get_value(key),
    "list": lambda pipe, key: pipe.lrange(key, 0, -1),
    "set": lambda pipe, key: pipe.smembers(key),
    "zset": lambda pipe, key: pipe.zrange(key, 0, -1),
    "hash": lambda pipe, key: pipe.hgetall(key),
}


PIPE_EXPECTED: dict[KeyType, Callable[[Any, int], bool]] = {
    "string": lambda res, ix: res == f"v{ix}",
    "list": lambda res, ix: set(res) == {f"v{ix}"},
    "set": lambda res, ix: res == {f"v{ix}"},
    "zset": lambda res, ix: res == [f"v{ix}"],
    "hash": lambda res, ix: res == {"value": f"v{ix}"},
}


PIPE_MISSING: dict[KeyType, Callable[[Any], bool]] = {
    "string": lambda res: res is None,
    "list": lambda res: res == [],
    "set": lambda res: res == set(),
    "zset": lambda res: res == [],
    "hash": lambda res: res == {},
}


@pytest.mark.parametrize("types", KEY_TYPE_REG)
@pytest.mark.parametrize("is_pipe", [False, True])
@pytest.mark.parametrize("rt_lua", [False, True])
def test_expire(
        types: KeyType,
        is_pipe: bool,
        rt_lua: bool) -> None:
    """
    Test expire command.

    Args:
        types (KeyType): Which key types to use.
        is_pipe (bool): Whether the operations are performed in a pipeline.
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    redis = Redis(rt=get_setup("test_expire", rt_lua))
    print(f"is_pipe={is_pipe} rt_lua={rt_lua} types={types}")
    if rt_lua and is_pipe:
        # FIXME: figure out why pipelines don't work correctly with redis
        return

    @contextmanager
    def block() -> Iterator[tuple[Redis | PipelineAPI, list[Action]]]:
        actions: list[Action] = []
        try:
            if is_pipe:
                with redis.pipeline() as pipe:
                    yield pipe, actions
                    res = pipe.execute()
                    assert len(res) == len(actions)
                    for val, (call, msg) in zip(res, actions):
                        assert call(val), msg(val)
            else:
                yield redis, actions
                assert len(actions) == 0
        finally:
            pass  # NOTE: making pylint happy

    def create(
            rt: Redis | PipelineAPI,
            actions: list[Action],
            key: str,
            ix: int) -> None:
        DEFAULTS[types](rt, key, ix)
        if is_pipe:
            actions.append((lambda _: True, lambda _: "?"))

    def expire(
            rt: Redis | PipelineAPI,
            actions: list[Action],
            key: str,
            *,
            expect: bool,
            mode: RExpireMode = REX_ALWAYS,
            expire_in: float | None = None,
            expire_timestamp: datetime | None = None) -> None:
        print(
            f"expire key {key} with {mode} in {expire_in or expire_timestamp}")
        if isinstance(rt, PipelineAPI):
            rt.expire(
                key,
                mode=mode,
                expire_in=expire_in,
                expire_timestamp=expire_timestamp)
            actions.append((
                lambda val: val == expect,
                lambda val: f"val={val} == expect={expect}"))
        else:
            assert expect == rt.expire(
                key,
                mode=mode,
                expire_in=expire_in,
                expire_timestamp=expire_timestamp)

    def ttl(
            rt: Redis | PipelineAPI,
            actions: list[Action],
            key: str,
            *,
            expect: bool | None) -> None:
        estr = None if expect is None else ("> 0.0" if expect else "<= 0.0")

        def check_ttl(res: float | None) -> bool:
            if expect is None:
                return res is None
            if res is None:
                return False
            if expect:
                return res > 0.0
            return res <= 0.0

        if isinstance(rt, PipelineAPI):
            rt.ttl(key)
            actions.append((
                check_ttl,
                lambda val: f"ttl of key {key} is {val} expected {estr}"))
        else:
            assert check_ttl(rt.ttl(key))

    def check(
            rt: Redis | PipelineAPI,
            actions: list[Action],
            key: str,
            ix: int | None) -> None:
        if ix is None:
            if isinstance(rt, PipelineAPI):
                PIPE_CHECKS[types](rt, key)
                actions.append(
                    (PIPE_MISSING[types], lambda val: f"{key} missing {val}"))
                rt.exists(key)
                actions.append(
                    (lambda val: val == 0, lambda val: f"exists {val} != 0"))
            else:
                assert MISSING[types](rt, key)
                assert rt.exists(key) == 0
            return
        if isinstance(rt, PipelineAPI):
            PIPE_CHECKS[types](rt, key)
            actions.append((
                lambda val: PIPE_EXPECTED[types](val, ix),
                lambda val: f"expected value with {key} and {ix} got {val}"))
        else:
            assert CHECKS[types](rt, key, ix)

    with block() as (rt, actions):
        create(rt, actions, "a", 1)
        create(rt, actions, "b", 2)
        create(rt, actions, "c", 3)
        create(rt, actions, "d", 4)
        create(rt, actions, "e", 5)
        create(rt, actions, "f", 6)
        create(rt, actions, "g", 7)
        create(rt, actions, "h", 8)
        create(rt, actions, "i", 9)
        create(rt, actions, "j", 10)
        create(rt, actions, "k", 11)

    expire_timestamp = now() + timedelta(seconds=0.3 if is_pipe else 0.1)
    with block() as (rt, actions):
        expire(rt, actions, "a", expect=False, mode=REX_EXPIRE, expire_in=0.1)
        expire(rt, actions, "b", expect=True, mode=REX_ALWAYS, expire_in=0.1)
        expire(
            rt,
            actions,
            "c",
            expect=True,
            mode=REX_PERSIST,
            expire_timestamp=expire_timestamp)
        expire(
            rt,
            actions,
            "d",
            expect=True,
            mode=REX_EARLIER,
            expire_timestamp=expire_timestamp)
        expire(
            rt,
            actions,
            "e",
            expect=False,
            mode=REX_LATER,
            expire_timestamp=expire_timestamp)
        expire(rt, actions, "f", expect=True, expire_in=0.3)
        expire(rt, actions, "g", expect=True, expire_in=0.3)
        expire(rt, actions, "h", expect=True, expire_in=1.0)
        expire(rt, actions, "i", expect=True, expire_in=0.3)
        expire(rt, actions, "j", expect=True, expire_in=1.0)
        expire(rt, actions, "k", expect=True, expire_in=0.5)

        check(rt, actions, "a", 1)
        check(rt, actions, "b", 2)
        check(rt, actions, "c", 3)
        check(rt, actions, "d", 4)
        check(rt, actions, "e", 5)
        check(rt, actions, "f", 6)
        check(rt, actions, "g", 7)
        check(rt, actions, "h", 8)
        check(rt, actions, "i", 9)
        check(rt, actions, "j", 10)
        check(rt, actions, "k", 11)
        if is_pipe:
            time.sleep(0.2)  # should execute before the pipe is executed

    with block() as (rt, actions):
        ttl(rt, actions, "a", expect=False)
        ttl(rt, actions, "b", expect=True)
        ttl(rt, actions, "c", expect=True)
        ttl(rt, actions, "d", expect=True)
        ttl(rt, actions, "e", expect=False)
        ttl(rt, actions, "f", expect=True)
        ttl(rt, actions, "g", expect=True)
        ttl(rt, actions, "h", expect=True)
        ttl(rt, actions, "i", expect=True)
        ttl(rt, actions, "j", expect=True)
        ttl(rt, actions, "k", expect=True)

    time.sleep(0.2)

    with block() as (rt, actions):
        check(rt, actions, "a", 1)
        check(rt, actions, "b", None)
        check(rt, actions, "c", None)
        check(rt, actions, "d", None)
        check(rt, actions, "e", 5)
        check(rt, actions, "f", 6)
        check(rt, actions, "g", 7)
        check(rt, actions, "h", 8)
        check(rt, actions, "i", 9)
        check(rt, actions, "j", 10)
        check(rt, actions, "k", 11)

        ttl(rt, actions, "a", expect=False)
        ttl(rt, actions, "b", expect=None)
        ttl(rt, actions, "c", expect=None)
        ttl(rt, actions, "d", expect=None)
        ttl(rt, actions, "e", expect=False)
        ttl(rt, actions, "f", expect=True)
        ttl(rt, actions, "g", expect=True)
        ttl(rt, actions, "h", expect=True)
        ttl(rt, actions, "i", expect=True)
        ttl(rt, actions, "j", expect=True)
        ttl(rt, actions, "k", expect=True)

    with block() as (rt, actions):
        expire(rt, actions, "a", expect=True, expire_in=0.1)
        expire(rt, actions, "b", expect=False, expire_in=0.1)
        expire(rt, actions, "e", expect=True, expire_in=0.0)
        expire(rt, actions, "f", expect=True, mode=REX_EXPIRE, expire_in=0.3)
        expire(rt, actions, "g", expect=True, mode=REX_LATER, expire_in=0.3)
        expire(rt, actions, "h", expect=True, mode=REX_EARLIER, expire_in=0.1)
        expire(rt, actions, "i", expect=True)
        expire(rt, actions, "j", expect=True, mode=REX_LATER)
        expire(rt, actions, "k", expect=False, mode=REX_EARLIER)

        check(rt, actions, "a", 1)
        check(rt, actions, "b", None)
        check(rt, actions, "c", None)
        check(rt, actions, "d", None)
        check(rt, actions, "e", None)
        check(rt, actions, "f", 6)
        check(rt, actions, "g", 7)
        check(rt, actions, "h", 8)
        check(rt, actions, "i", 9)
        check(rt, actions, "j", 10)
        check(rt, actions, "k", 11)

        ttl(rt, actions, "a", expect=True)
        ttl(rt, actions, "b", expect=None)
        ttl(rt, actions, "c", expect=None)
        ttl(rt, actions, "d", expect=None)
        ttl(rt, actions, "e", expect=None)
        ttl(rt, actions, "f", expect=True)
        ttl(rt, actions, "g", expect=True)
        ttl(rt, actions, "h", expect=True)
        ttl(rt, actions, "i", expect=False)
        ttl(rt, actions, "j", expect=False)
        ttl(rt, actions, "k", expect=True)

    time.sleep(0.2)

    with block() as (rt, actions):
        check(rt, actions, "a", None)
        check(rt, actions, "b", None)
        check(rt, actions, "c", None)
        check(rt, actions, "d", None)
        check(rt, actions, "e", None)
        check(rt, actions, "f", 6)
        check(rt, actions, "g", 7)
        check(rt, actions, "h", None)
        check(rt, actions, "i", 9)
        check(rt, actions, "j", 10)
        check(rt, actions, "k", 11)

        ttl(rt, actions, "a", expect=None)
        ttl(rt, actions, "b", expect=None)
        ttl(rt, actions, "c", expect=None)
        ttl(rt, actions, "d", expect=None)
        ttl(rt, actions, "e", expect=None)
        ttl(rt, actions, "f", expect=True)
        ttl(rt, actions, "g", expect=True)
        ttl(rt, actions, "h", expect=None)
        ttl(rt, actions, "i", expect=False)
        ttl(rt, actions, "j", expect=False)
        ttl(rt, actions, "k", expect=True)

    time.sleep(0.2)

    with block() as (rt, actions):
        check(rt, actions, "a", None)
        check(rt, actions, "b", None)
        check(rt, actions, "c", None)
        check(rt, actions, "d", None)
        check(rt, actions, "e", None)
        check(rt, actions, "f", None)
        check(rt, actions, "g", None)
        check(rt, actions, "h", None)
        check(rt, actions, "i", 9)
        check(rt, actions, "j", 10)
        check(rt, actions, "k", None)

        ttl(rt, actions, "a", expect=None)
        ttl(rt, actions, "b", expect=None)
        ttl(rt, actions, "c", expect=None)
        ttl(rt, actions, "d", expect=None)
        ttl(rt, actions, "e", expect=None)
        ttl(rt, actions, "f", expect=None)
        ttl(rt, actions, "g", expect=None)
        ttl(rt, actions, "h", expect=None)
        ttl(rt, actions, "i", expect=False)
        ttl(rt, actions, "j", expect=False)
        ttl(rt, actions, "k", expect=None)

    # FIXME: test calling expire from a script
